#!/usr/bin/env python
# jsn.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (jsn.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: jsn.

"""

#------------------------------------------------------------------------------

import sys
import json

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

from bitdust.lib import strng

#------------------------------------------------------------------------------


def dict_keys_to_text(dct, encoding='utf-8', errors='strict'):
    """
    Returns dict where all keys are converted to text strings.
    Only works for keys in a "root" level of the dict.
    """
    return {(k.decode(encoding, errors=errors) if strng.is_bin(k) else k): v for k, v in dct.items()}


def dict_keys_to_bin(dct, encoding='utf-8', errors='strict'):
    """
    Returns dict where all keys are converted to binary strings.
    Only works for keys in a "root" level of the dict.
    """
    return {(k.encode(encoding, errors=errors) if strng.is_text(k) else k): v for k, v in dct.items()}


def dict_values_to_text(dct, encoding='utf-8', errors='strict'):
    """
    Returns dict where all values are converted to text strings.
    Can go recursively, but not super smart.
    If value is a list of dicts - will not be converted.
    """
    # TODO: make it fully recursive... for example if list of lists is passed
    _d = {}
    for k, v in dct.items():
        _v = v
        if strng.is_bin(_v):
            _v = _v.decode(encoding, errors=errors)
        elif isinstance(_v, dict):
            _v = dict_values_to_text(_v, encoding=encoding, errors=errors)
        elif isinstance(_v, list):
            _v = [i.decode(encoding, errors=errors) if strng.is_bin(i) else i for i in _v]
        elif isinstance(_v, tuple):
            _v = tuple([i.decode(encoding, errors=errors) if strng.is_bin(i) else i for i in _v])
        _d[k] = _v
    return _d


def dict_items_to_text(dct, encoding='utf-8', errors='strict'):
    """
    Returns dict where all keys and values are converted to text strings.
    Only works for simple dicts - one level structure.
    """
    # TODO: make it fully recursive... for example if list of lists is passed
    if strng.is_text(dct):
        return dct
    elif strng.is_bin(dct):
        return dct.decode(encoding, errors=errors)
    elif isinstance(dct, list):
        return [dict_items_to_text(i) for i in dct]
    elif isinstance(dct, tuple):
        return tuple([dict_items_to_text(i) for i in dct])
    elif isinstance(dct, dict):
        return {strng.to_text(x): dict_items_to_text(y) for x, y in dct.items()}
    if not isinstance(dct, dict):
        return dct
    _d = {}
    for k in dct.keys():
        _v = dct[k]
        if strng.is_text(_v):
            pass
        elif strng.is_bin(_v):
            _v = _v.decode(encoding, errors=errors)
        elif isinstance(_v, list):
            _v = [dict_items_to_text(i) for i in _v]
        elif isinstance(_v, tuple):
            _v = tuple([dict_items_to_text(i) for i in _v])
        elif isinstance(_v, dict):
            _v = {strng.to_text(x): dict_items_to_text(y) for x, y in _v.items()}
        _k = k
        if strng.is_bin(_k):
            _k = _k.decode(encoding, errors=errors)
        _d[_k] = _v
    return _d


#------------------------------------------------------------------------------


def pack_dict(dct, encoding='utf-8', errors='strict'):
    """
    Creates another dict from input dict where types of keys and values are also present.
    Keys can only be bin/text strings, integers, floats or None.
    Values can only be bin/text strings, integers, floats, None, lists or tuples.
    Result dict will always contain only text (unicode) keys and values or simple types like integer, float or None.
    """
    if not dct:
        return {}
    _d = {}
    for k, v in dct.items():
        _k = k
        _ktyp = 's'
        if strng.is_bin(_k):
            _k = _k.decode(encoding, errors=errors)
            _ktyp = 'b'
        elif isinstance(_k, int):
            _ktyp = 'i'
        elif isinstance(_k, float):
            _ktyp = 'f'
        elif _k is None:
            _ktyp = 'n'
        _v = v
        _vtyp = 's'
        if strng.is_bin(_v):
            _v = _v.decode(encoding, errors=errors)
            _vtyp = 'b'
        elif isinstance(_v, int):
            _vtyp = 'i'
        elif isinstance(_v, float):
            _vtyp = 'f'
        elif isinstance(_v, dict):
            _v = pack_dict(_v, encoding=encoding, errors=errors)
            _vtyp = 'd'
        elif isinstance(_v, list):
            _v = [pack_dict({'i': i}, encoding=encoding, errors=errors) for i in _v]
            _vtyp = 'l'
        elif isinstance(_v, tuple):
            _v = [pack_dict({'i': i}, encoding=encoding, errors=errors) for i in _v]
            _vtyp = 't'
        elif _v is None:
            _vtyp = 'n'
        _d[_k] = (
            _ktyp,
            _vtyp,
            _v,
        )
    return _d


def unpack_dict(dct, encoding='utf-8', errors='strict'):
    """
    Reverse operation of `pack_dict()` method - returns original dict with all keys and values of correct types.
    """
    if not dct:
        return {}
    _d = {}
    for k, v in dct.items():
        _k = k
        if len(v) != 3:
            raise ValueError('unpack failed, invalid value: %r' % v)
        if v[0] == 'b':
            _k = _k.encode(encoding, errors=errors)
        _v = v[2]
        if v[1] == 'b':
            _v = _v.encode(encoding, errors=errors)
        elif v[1] == 'd':
            _v = unpack_dict(_v, encoding=encoding, errors=errors)
        elif v[1] == 'l':
            _v = [unpack_dict(i, encoding=encoding, errors=errors)['i'] for i in _v]
        elif v[1] == 't':
            _v = tuple([unpack_dict(i, encoding=encoding, errors=errors)['i'] for i in _v])
        _d[_k] = _v
    return _d


#------------------------------------------------------------------------------


def dumps(obj, indent=None, separators=None, sort_keys=None, ensure_ascii=False, encoding='utf-8', keys_to_text=False, values_to_text=False, empty_result='{}', **kw):
    """
    Calls `json.dumps()` with parameters.
    Always translates every byte string json value into text using encoding.
    """
    if obj is None or obj == '' or obj == b'':
        return empty_result

    enc_errors = kw.pop('errors', 'strict')

    def _to_text(v):
        if strng.is_bin(v):
            v = v.decode(encoding, errors=enc_errors)
        if not strng.is_text(v):
            v = strng.to_text(v)
        return v

    if keys_to_text:
        obj = dict_keys_to_text(obj, encoding=encoding, errors=enc_errors)

    if values_to_text:
        obj = dict_values_to_text(obj, encoding=encoding, errors=enc_errors)

    try:
        if sys.version_info[0] < 3:
            return json.dumps(obj=obj, indent=indent, separators=separators, sort_keys=sort_keys, ensure_ascii=ensure_ascii, default=_to_text, encoding=encoding, **kw)
        else:
            return json.dumps(obj=obj, indent=indent, separators=separators, sort_keys=sort_keys, ensure_ascii=ensure_ascii, default=_to_text, **kw)
    except Exception as exc:
        if _Debug:
            import os
            import tempfile
            fd, _ = tempfile.mkstemp(suffix='err', prefix='jsn_dumps_', text=True)
            try:
                os.write(fd, repr(obj))
            except:
                try:
                    os.write(fd, strng.to_bin(repr(type(obj))))
                except:
                    os.write(fd, b'failed to serialize object')
            os.close(fd)
        raise exc


#------------------------------------------------------------------------------


def loads(s, encoding='utf-8', keys_to_bin=False, **kw):
    """
    Calls `json.loads()` with parameters.
    Always translates all json values into binary strings using encoding.
    Also do the same for keys when `keys_to_bin=True`.
    """
    if not s:
        return None

    def _to_bin(dct):
        for k in dct.keys():
            if strng.is_text(dct[k]):
                dct[k] = dct[k].encode(encoding)
        if keys_to_bin:
            return {(k.encode(encoding) if strng.is_text(k) else k): v for k, v in dct.items()}
        return dct

    try:
        return json.loads(s=s, object_hook=_to_bin, **kw)
    except Exception as exc:
        if _Debug:
            try:
                import os
                import tempfile
                fd, _ = tempfile.mkstemp(suffix='err', prefix='jsn_loads_', text=True)
                os.write(fd, s)
                os.close(fd)
            except:
                pass
        raise exc


#------------------------------------------------------------------------------


def loads_text(s, encoding='utf-8', **kw):
    """
    Calls `json.loads()` with parameters.
    Always translates all json keys and values into unicode strings.
    """
    if not s:
        return None

    enc_errors = kw.pop('errors', 'strict')

    try:
        return json.loads(s=s, object_hook=lambda itm: dict_items_to_text(itm, encoding=encoding, errors=enc_errors), **kw)
    except Exception as exc:
        if _Debug:
            import os
            import tempfile
            fd, _ = tempfile.mkstemp(suffix='err', prefix='jsn_loads_', text=True)
            os.write(fd, s)
            os.close(fd)
        raise exc



# optipoolware.py v 0.397 to be used with Python3.5 or later
# Bismuth pool mining software
# Copyright Hclivess, Maccaspacca 2017, 2018
# for license see LICENSE file
# .

import socketserver, connections, time, options, log, sqlite3, socks, random, re, essentials, base64, sys, os, math
from Cryptodome.Hash import SHA
from Cryptodome.Signature import PKCS1_v1_5
from Cryptodome.PublicKey import RSA
import threading
import json

import logging
import mining_heavy3 as mining

config = options.Get()
config.read()
port = config.port
node_ip_conf = config.node_ip
ledger_path_conf = 'static/ledger.db'
#tor_conf = config.tor
debug_level_conf = config.debug_level
version = config.version

if version == 'testnet':
    port = '2829'
    m_peer_file = 'peers_test.txt'
    ledger_path_conf = 'static/test.db'
else:
    m_peer_file = 'peers.txt'

print('Peers file: {}'.format(m_peer_file))

# print(version)

# load config

#key, public_key_readable, private_key_readable, _, _, public_key_hashed, address = essentials.keys_load ("privkey.der", "pubkey.der")
key, public_key_readable, private_key_readable, encrypted, unlocked, public_key_hashed, address, keyfile = essentials.keys_load('privkey.der', 'pubkey.der')

app_log = log.log('pool.log', debug_level_conf)

# This part is what goes on console.
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s %(funcName)s(%(lineno)d) %(message)s')
ch.setFormatter(formatter)
app_log.addHandler(ch)
new_time = 0
new_diff = 0

print('Pool Address: {}'.format(address))

# load config
try:

    lines = [line.rstrip('\n') for line in open('pool.txt')]
    for line in lines:
        try:
            if 'mine_diff=' in line:
                mdiff = int(line.split('=')[1])
        except Exception as e:
            mdiff = 65
        try:
            if 'min_payout=' in line:
                min_payout = float(line.split('=')[1])
        except Exception as e:
            min_payout = 1
        try:
            if 'pool_fee=' in line:
                pool_fee = float(line.split('=')[1])
        except Exception as e:
            pool_fee = 0
        try:
            if 'alt_fee=' in line:
                alt_fee = float(line.split('=')[1])
        except Exception as e:
            alt_fee = 0
        try:
            if 'worker_time=' in line:
                w_time = int(line.split('=')[1])
        except Exception as e:
            w_time = 10
        try:
            if 'alt_add=' in line:
                alt_add = str(line.split('=')[1])
        except Exception as e:
            alt_add = '1aae2cfe5d01acc8d7cbc90fcf8bb715ca24927504d0d8071c0979c7'

except Exception as e:
    min_payout = 1
    mdiff = 65
    pool_fee = 0
    alt_fee = 0
    w_time = 10
    alt_add = '1aae2cfe5d01acc8d7cbc90fcf8bb715ca24927504d0d8071c0979c7'
# load config

#m = socks.socksocket()
#m.connect((node_ip_conf, int(port)))  # connect to local node
#connections.send(m, "api_mempool", 10)
#tresult = connections.receive(m, 10)
#m.close()

#print(tresult)

bin_format_dict = dict((x, format(ord(x), '8b').replace(' ', '0')) for x in '0123456789abcdef')


def percentage(percent, whole):
    return int((percent*whole)/100)


def checkdb():
    shares = sqlite3.connect('shares.db')
    shares.text_factory = str
    s = shares.cursor()
    s.execute('SELECT * FROM shares')
    present = s.fetchall()

    if not present:
        return False
    else:
        return True


# payout processing


def payout(payout_threshold, myfee, othfee):
    global node_ip_conf
    global port

    print('Minimum payout is {} Bismuth'.format(str(payout_threshold)))
    print('Current pool fee is {} Percent'.format(str(myfee)))

    shares = sqlite3.connect('shares.db')
    shares.text_factory = str
    s = shares.cursor()

    conn = sqlite3.connect(ledger_path_conf, timeout=60.0)
    conn.text_factory = str
    c = conn.cursor()

    #get sum of all shares not paid
    s.execute('SELECT sum(shares) FROM shares WHERE paid != 1')
    shares_total = s.fetchone()[0]
    #get sum of all shares not paid

    #get block threshold
    try:
        s.execute('SELECT min(timestamp) FROM shares WHERE paid != 1')
        block_threshold = float(s.fetchone()[0])
    except:
        block_threshold = time.time()
    #get block threshold

    #get eligible blocks
    reward_list = []
    for row in c.execute('SELECT * FROM transactions WHERE address = ? AND CAST(timestamp AS INTEGER) >= ? AND reward != 0', (address, ) + (block_threshold, )):
        reward_list.append(float(row[9]))

    super_total = sum(reward_list)
    #get eligible blocks

    # so now we have sum of shares, total reward, block threshold

    # reduce total rewards by total fees percentage
    reward_total = '%.8f' % (((100 - (myfee + othfee))*super_total)/100)
    reward_total = float(reward_total)

    if reward_total > 0:

        # calculate alt address fee

        ft = super_total - reward_total
        try:
            at = '%.8f' % (ft*(othfee/(myfee + othfee)))
        except:
            at = 0

        # calculate reward per share
        reward_per_share = reward_total/shares_total

        # calculate shares threshold for payment

        shares_threshold = math.floor(payout_threshold/reward_per_share)

        #get unique addresses
        addresses = []
        for row in s.execute('SELECT * FROM shares'):
            shares_address = row[0]

            if shares_address not in addresses:
                addresses.append(shares_address)
        print('payout address', addresses)
        #get unique addresses

        # prepare payout address list with number of shares and new total shares
        payadd = []
        new_sum = 0
        for x in addresses:
            s.execute('SELECT sum(shares) FROM shares WHERE address = ? AND paid != 1', (x, ))
            shares_sum = s.fetchone()[0]

            if shares_sum == None:
                shares_sum = 0
            if shares_sum > shares_threshold:
                payadd.append([x, shares_sum])
                new_sum = new_sum + shares_sum
        #prepare payout address list with number of shares and new total shares

        # recalculate reward per share now we have removed those below payout threshold
        try:

            reward_per_share = reward_total/new_sum

        except:
            reward_per_share = 0

        print('reward_per_share', reward_per_share)

        paylist = []
        for p in payadd:
            payme = '%.8f' % (p[1]*reward_per_share)
            paylist.append([p[0], payme])

        if othfee > 0:
            paylist.append([alt_add, at])

        payout_passed = 0
        for r in paylist:
            print(r)
            recipient = r[0]
            claim = float(r[1])

            payout_passed = 1
            openfield = 'pool'
            keep = 0
            fee = float('%.8f' % float(0.01 + (float(len(openfield))/100000) + (keep)))  # 0.01 + openfield fee + keep fee
            #make payout

            timestamp = '%.2f' % time.time()
            transaction = (str(timestamp), str(address), str(recipient), '%.8f' % float(claim - fee), str(keep), str(openfield))  # this is signed
            # print transaction

            h = SHA.new(str(transaction).encode('utf-8'))
            signer = PKCS1_v1_5.new(key)
            signature = signer.sign(h)
            signature_enc = base64.b64encode(signature)
            print('Encoded Signature: {}'.format(signature_enc.decode('utf-8')))

            verifier = PKCS1_v1_5.new(key)
            if verifier.verify(h, signature) == True:
                print('The signature is valid, proceeding to send transaction')
                txid = signature_enc[:56]
                mytxid = txid.decode('utf-8')
                tx_submit = (str(timestamp), str(address), str(recipient), '%.8f' % float(claim - fee), str(signature_enc.decode('utf-8')), str(public_key_hashed.decode('utf-8')), str(keep), str(openfield))  #float kept for compatibility

                t = socks.socksocket()
                t.connect((node_ip_conf, int(port)))  # connect to local node

                connections.send(t, 'mpinsert', 10)
                connections.send(t, [tx_submit], 10)
                reply = connections.receive(t, 10)
            else:
                print('Invalid signature')
                reply = 'Invalid signature'

                print('Transaction sent with txid: {}'.format(mytxid))

            t.close()

            s.execute('UPDATE shares SET paid = 1 WHERE address = ?', (recipient, ))
            shares.commit()

        if payout_passed == 1:
            s.execute('UPDATE shares SET timestamp = ?', (time.time(), ))
            shares.commit()

        # calculate payouts
        #payout

        # archive paid shares
        s.execute('SELECT * FROM shares WHERE paid = 1')
        pd = s.fetchall()

        if pd == None:
            pass
        else:
            archive = sqlite3.connect('archive.db')
            archive.text_factory = str
            a = archive.cursor()

            for sh in pd:
                a.execute('INSERT INTO shares VALUES (?,?,?,?,?,?,?,?)', (sh[0], sh[1], sh[2], sh[3], sh[4], sh[5], sh[6], sh[7]))

            archive.commit()
            a.close()
        # archive paid shares

    # clear nonces
    s.execute('DELETE FROM nonces')
    s.execute('DELETE FROM shares WHERE paid = 1')
    shares.commit()
    s.execute('VACUUM')
    #clear nonces
    s.close()


def commit(cursor):
    # secure commit for slow nodes
    passed = 0
    while passed == 0:
        try:
            cursor.commit()
            passed = 1
        except Exception as e:
            app_log.warning('Retrying database execute due to ' + str(e))
            time.sleep(random.random())
            # secure commit for slow nodes


def execute(cursor, what):
    # secure execute for slow nodes
    passed = 0
    while passed == 0:
        try:
            # print cursor
            # print what

            cursor.execute(what)
            passed = 1
        except Exception as e:
            app_log.warning('Retrying database execute due to {}'.format(e))
            time.sleep(random.random())
            # secure execute for slow nodes
    return cursor


def execute_param(cursor, what, param):
    # secure execute for slow nodes
    passed = 0
    while passed == 0:
        try:
            # print cursor
            # print what
            cursor.execute(what, param)
            passed = 1
        except Exception as e:
            app_log.warning('Retrying database execute due to ' + str(e))
            time.sleep(0.1)
            # secure execute for slow nodes
    return cursor


def bin_convert(string):
    return ''.join(bin_format_dict[x] for x in string)


def bin_convert_orig(string):
    return ''.join(format(ord(x), '8b').replace(' ', '0') for x in string)


def s_test(testString):

    if testString.isalnum():
        if (re.search('[abcdef]', testString)):
            if len(testString) == 56:
                return True
    else:
        return False


def n_test(testString):

    if testString.isalnum():
        if (re.search('[abcdef]', testString)):
            if len(testString) < 129:
                return True
    else:
        return False


def paydb():
    global new_time
    while True:
        app_log.warning('Payout run finished')
        time.sleep(3601)
        #time.sleep(60) # test
        v = float('%.2f' % time.time())
        v1 = new_time
        v2 = v - v1

        if v2 < 100000:
            payout(min_payout, pool_fee, alt_fee)
            app_log.warning('Payout running...')
        else:
            app_log.warning('Node over 1 mins out: %r ...payout delayed' % v2)


def worker(s_time):
    global new_diff
    global new_hash
    global new_time
    doclean = 0

    n = socks.socksocket()
    n.connect((node_ip_conf, int(port)))  # connect to local node

    while True:

        time.sleep(s_time)
        #doclean +=1

        try:

            app_log.warning('Worker task...')
            connections.send(n, 'blocklast', 10)
            blocklast = connections.receive(n, 10)

            connections.send(n, 'diffget', 10)
            diff = connections.receive(n, 10)

            new_hash = blocklast[7]
            new_time = blocklast[1]
            new_diff = math.floor(diff[1])

            app_log.warning('Difficulty = {}'.format(str(new_diff)))
            app_log.warning('Blockhash = {}'.format(str(new_hash)))
            # print("Worker")

        except Exception as e:
            app_log.warning(str(e))
    n.close()


# TODO: for tests only
#os.remove('shares.db')
#os.remove('archive.db')

if not os.path.exists('shares.db'):
    # create empty shares
    shares = sqlite3.connect('shares.db')
    shares.text_factory = str
    s = shares.cursor()
    execute(s, 'CREATE TABLE IF NOT EXISTS shares (address, shares, timestamp, paid, rate, name, workers, subname)')
    execute(s, 'CREATE TABLE IF NOT EXISTS nonces (nonce)')  #for used hash storage
    app_log.warning('Created shares file')
    s.close()
    # create empty shares
if not os.path.exists('archive.db'):
    # create empty archive
    archive = sqlite3.connect('archive.db')
    archive.text_factory = str
    a = archive.cursor()
    execute(a, 'CREATE TABLE IF NOT EXISTS shares (address, shares, timestamp, paid, rate, name, workers, subname)')
    app_log.warning('Created archive file')
    a.close()
    # create empty archive
"""
if checkdb():
    payout(min_payout,pool_fee,alt_fee)
"""


class MyTCPHandler(socketserver.BaseRequestHandler):
    def handle(self):
        global new_diff
        key = RSA.importKey(private_key_readable)

        self.allow_reuse_address = True

        peer_ip = self.request.getpeername()[0]

        try:
            data = connections.receive(self.request, 10)

            app_log.warning('Received: {} from {}'.format(data, peer_ip))  # will add custom ports later

            if data == 'getwork':  # sends the miner the blockhash and mining diff for shares

                work_send = []
                work_send.append((new_hash, mdiff, address, mdiff))

                connections.send(self.request, work_send, 10)

                print('Work package sent.... {}'.format(str(new_hash)))

            elif data == 'block':  # from miner to node

                # sock
                #s1 = socks.socksocket()
                #if tor_conf == 1:
                #	s1.setproxy(socks.PROXY_TYPE_SOCKS5, "127.0.0.1", 9050)
                #s1.connect(("127.0.0.1", int(port)))  # connect to local node,
                # sock

                # receive nonce from miner
                miner_address = connections.receive(self.request, 10)

                if not s_test(miner_address):

                    app_log.warning('Bad Miner Address Detected - Changing to default')
                    miner_address = alt_add
                    #s1.close()

                else:

                    app_log.warning('Received a solution from miner {} ({})'.format(peer_ip, miner_address))

                    block_nonce = connections.receive(self.request, 10)
                    block_timestamp = (block_nonce[-1][0])
                    nonce = (block_nonce[-1][1])
                    mine_hash = ((block_nonce[-1][2]))  # block hash claimed
                    ndiff = ((block_nonce[-1][3]))  # network diff when mined
                    sdiffs = ((block_nonce[-1][4]))  # actual diff mined
                    mrate = ((block_nonce[-1][5]))  # total hash rate in khs
                    bname = ((block_nonce[-1][6]))  # base worker name
                    wnum = ((block_nonce[-1][7]))  # workers
                    wstr = ((block_nonce[-1][8]))  # worker number
                    wname = '{}{}'.format(bname, wstr)  # worker name

                    app_log.warning('Mined nonce details: {}'.format(block_nonce))
                    app_log.warning('Claimed hash: {}'.format(mine_hash))
                    app_log.warning('Claimed diff: {}'.format(sdiffs))

                    if not n_test(nonce):
                        app_log.warning('Bad Nonce Format Detected - Closing Connection')
                        self.close
                    app_log.warning('Processing nonce.....')

                    diff = new_diff
                    db_block_hash = mine_hash

                    real_diff = mining.diffme_heavy3(address, nonce, db_block_hash)
                    """
                    mining_hash = bin_convert_orig(hashlib.sha224((address + nonce + db_block_hash).encode("utf-8")).hexdigest())
                    mining_condition = bin_convert_orig(db_block_hash)[0:diff]

                    if mining_condition in mining_hash:
                    """
                    app_log.warning('Solution has {} difficulty, current difficulty is {}'.format(real_diff, diff))
                    if real_diff >= int(diff):

                        app_log.warning('Difficulty requirement satisfied for mining')
                        app_log.warning('Sending block to nodes')

                        cn = options.Get()
                        cn.read()
                        cport = cn.port
                        try:
                            cnode_ip_conf = cn.node_ip_conf
                        except:
                            cnode_ip_conf = cn.node_ip

                        #ctor_conf = cn.tor_conf
                        cversion = cn.version

                        if cversion == 'testnet':
                            cport = '2829'
                            m_peer_file = 'peers_test.txt'
                        else:
                            m_peer_file = 'peers.txt'

                        app_log.warning('Local node ip {} on port {}'.format(cnode_ip_conf, cport))

                        m = socks.socksocket()
                        m.connect((cnode_ip_conf, int(cport)))  # connect to local node
                        connections.send(m, 'api_mempool', 10)
                        result = connections.receive(m, 10)
                        app_log.warning('I have got to receive mempool')
                        m.close()

                        # include data
                        block_send = []
                        del block_send[:]  # empty
                        removal_signature = []
                        del removal_signature[:]  # empty

                        app_log.warning('prepare empty block and clear data')

                        for dbdata in result:
                            transaction = (str(dbdata[0]), str(dbdata[1][:56]), str(dbdata[2][:56]), '%.8f' % float(dbdata[3]), str(dbdata[4]), str(dbdata[5]), str(dbdata[6]), str(dbdata[7]))  # create tuple
                            block_send.append(transaction)  # append tuple to list for each run
                            removal_signature.append(str(dbdata[4]))  # for removal after successful mining

                        # claim reward
                        transaction_reward = tuple
                        transaction_reward = (str(block_timestamp), str(address[:56]), str(address[:56]), '%.8f' % float(0), '0', str(nonce))  # only this part is signed!
                        print('transaction_reward', transaction_reward)

                        h = SHA.new(str(transaction_reward).encode('utf-8'))
                        signer = PKCS1_v1_5.new(key)
                        signature = signer.sign(h)
                        signature_enc = base64.b64encode(signature)

                        if signer.verify(h, signature) == True:
                            app_log.warning('Signature valid')

                            block_send.append((str(block_timestamp), str(address[:56]), str(address[:56]), '%.8f' % float(0), str(signature_enc.decode('utf-8')), str(public_key_hashed.decode('utf-8')), '0', str(nonce)))  # mining reward tx
                            # app_log.warning('Block to send: {}'.format(block_send))

                            if not any(isinstance(el, list) for el in block_send):  # if it's not a list of lists (only the mining tx and no others)
                                new_list = []
                                new_list.append(block_send)
                                block_send = new_list  # make it a list of lists
                                app_log.warning(block_send)

                        global peer_dict
                        peer_dict = {}

                        with open(m_peer_file) as f:
                            peer_dict = json.load(f)

                            app_log.warning(peer_dict)

                            for k, v in peer_dict.items():
                                peer_ip = k
                                # app_log.info(HOST)
                                peer_port = int(v)
                                # app_log.info(PORT)
                                # connect to all nodes

                                try:
                                    s = socks.socksocket()
                                    s.settimeout(0.3)
                                    #if ctor_conf == 1:
                                    #    s.setproxy(socks.PROXY_TYPE_SOCKS5, "127.0.0.1", 9050)
                                    s.connect((peer_ip, int(peer_port)))  # connect to node in peerlist
                                    app_log.warning('Connected')

                                    app_log.warning('Miner: Proceeding to submit mined block')

                                    connections.send(s, 'block', 10)
                                    #connections.send(s, address, 10)
                                    connections.send(s, block_send, 10)

                                    app_log.warning('Miner: Block submitted to {}'.format(peer_ip))
                                except Exception as e:
                                    app_log.warning('Miner: Could not submit block to {} because {}'.format(peer_ip, e))

                    if diff < mdiff:
                        diff_shares = diff
                    else:
                        diff_shares = mdiff

                    shares = sqlite3.connect('shares.db')
                    shares.text_factory = str
                    s = shares.cursor()

                    # protect against used share resubmission
                    execute_param(s, ('SELECT nonce FROM nonces WHERE nonce = ?'), (nonce, ))

                    try:
                        result = s.fetchone()[0]
                        app_log.warning('Miner trying to reuse a share, ignored')
                    except:
                        # protect against used share resubmission
                        """
                        mining_condition = bin_convert_orig(db_block_hash)[0:diff_shares] #floor set by pool
                        if mining_condition in mining_hash:
                        """
                        if real_diff >= diff_shares:
                            app_log.warning('Difficulty requirement satisfied for saving shares \n')

                            execute_param(s, ('INSERT INTO nonces VALUES (?)'), (nonce, ))
                            commit(shares)

                            timestamp = '%.2f' % time.time()

                            s.execute('INSERT INTO shares VALUES (?,?,?,?,?,?,?,?)', (str(miner_address), str(1), timestamp, '0', str(mrate), bname, str(wnum), wname))
                            shares.commit()

                        else:
                            app_log.warning('Difficulty requirement not satisfied for anything \n')

                    s.close()

            self.request.close()
        except Exception as e:
            app_log.error('Error: {}'.format(e))

    app_log.warning('Starting up...')


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


if __name__ == '__main__':

    mining.mining_open()

    try:
        # Disabling payouts
        # background_thread = threading.Thread(target=paydb)
        # background_thread.daemon = True
        # background_thread.start()

        worker_thread = threading.Thread(target=worker, args=(w_time, ))
        worker_thread.daemon = True
        worker_thread.start()
        app_log.warning('Starting up background tasks....')
        time.sleep(10)

        try:
            pool_port = int(sys.argv[1])
        except Exception as e:
            pool_port = 8525

        HOST, PORT = '0.0.0.0', pool_port

        # Create the server thread handler, binding to localhost on port above
        server = ThreadedTCPServer((HOST, PORT), MyTCPHandler)
        ip, server_port = server.server_address

        server_thread = threading.Thread(target=server.serve_forever)

        server_thread.daemon = True
        server_thread.start()
        server_thread.join()
        server.shutdown()
        server.server_close()
    finally:
        mining.mining_close()



#!/usr/bin/env python
# postman.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (postman.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

import os
import sys

#------------------------------------------------------------------------------

try:
    pass
except:
    sys.exit('Error initializing twisted.internet.reactor in keys_synchronizer.py')

#------------------------------------------------------------------------------

from twisted.internet.defer import DeferredList

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.crypt import my_keys

from bitdust.lib import jsn
from bitdust.lib import utime
from bitdust.lib import packetid

from bitdust.contacts import identitycache

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.main import settings
from bitdust.main import events

from bitdust.p2p import p2p_service
from bitdust.p2p import propagate

from bitdust.stream import p2p_queue
from bitdust.stream import message

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_ActiveStreams = {}
_ActiveCustomers = {}

#------------------------------------------------------------------------------


def streams():
    global _ActiveStreams
    return _ActiveStreams


def customers():
    global _ActiveCustomers
    return _ActiveCustomers


#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'postman.init')
    events.add_subscriber(on_identity_url_changed, 'identity-url-changed')
    message.consume_messages(
        consumer_callback_id='postman',
        callback=on_consume_queue_messages,
        direction='incoming',
        message_types=[
            'queue_message',
        ],
    )
    p2p_queue.add_message_processed_callback(on_message_processed)
    start_all_streams()
    do_cache_known_customers()


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'postman.shutdown')
    stop_all_streams()
    p2p_queue.remove_message_processed_callback(on_message_processed)
    message.clear_consumer_callbacks('postman')
    events.remove_subscriber(on_identity_url_changed, 'identity-url-changed')


#------------------------------------------------------------------------------


def register_stream(queue_id):
    if queue_id in streams():
        raise Exception('queue already exist')
    queue_info = global_id.ParseGlobalQueueID(queue_id)
    customer_idurl = global_id.glob2idurl(queue_info['owner_id'])
    if not customer_idurl:
        raise Exception('unknown customer')
    if not id_url.is_cached(customer_idurl):
        raise Exception('customer idurl %r is not cached yet' % customer_idurl)
    if customer_idurl not in customers():
        customers()[customer_idurl] = []
    else:
        if queue_id in customers()[customer_idurl]:
            raise Exception('queue is already registered for that customer')
    customers()[customer_idurl].append(queue_id)
    streams()[queue_id] = {
        'active': False,
        'consumers': {},
        'producers': {},
        'last_sequence_id': -1,
    }
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, customer=customer_idurl, customer_streams=len(customers().get(customer_idurl, [])))
    return True


def unregister_stream(queue_id):
    if queue_id not in streams():
        raise Exception('queue is not exist')
    queue_info = global_id.ParseGlobalQueueID(queue_id)
    customer_idurl = global_id.glob2idurl(queue_info['owner_id'])
    if not customer_idurl:
        raise Exception('unknown customer')
    if not id_url.is_cached(customer_idurl):
        raise Exception('customer idurl %r is not cached yet' % customer_idurl)
    if customer_idurl not in customers():
        raise Exception('customer is not registered')
    if queue_id not in customers()[customer_idurl]:
        raise Exception('queue is not registered for that customer')
    customers()[customer_idurl].remove(queue_id)
    if not customers()[customer_idurl]:
        customers().pop(customer_idurl)
    streams().pop(queue_id)
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, customer=customer_idurl, customer_streams=len(customers().get(customer_idurl, [])))
    return True


#------------------------------------------------------------------------------


def on_consume_queue_messages(json_messages):
    received = 0
    pushed = 0
    handled = False
    for json_message in json_messages:
        try:
            msg_type = json_message.get('type', '')
            msg_direction = json_message['dir']
            packet_id = json_message['packet_id']
            from_idurl = json_message['owner_idurl']
            to_user = json_message['to']
            msg_data = json_message['data']
            msg_action = msg_data.get('action', 'read')
        except:
            lg.exc()
            continue
        if msg_direction != 'incoming':
            continue
        if global_id.latest_glob_id(to_user) != my_id.getID():
            continue
        if msg_type != 'queue_message':
            continue
        if msg_action not in ['produce', 'consume']:
            continue
        queue_id = msg_data.get('queue_id')
        if not queue_id:
            continue
        if msg_type == 'queue_message':
            if queue_id not in streams():
                group_creator_idurl = global_id.GetGlobalQueueOwnerIDURL(queue_id)
                if not group_creator_idurl.is_latest():
                    lg.warn('group creator idurl was rotated, consumer must refresh own identity cache: %r ~ %r' % (group_creator_idurl.to_original(), group_creator_idurl.to_bin()))
                    known_ident = identitycache.get_one(group_creator_idurl.to_bin())
                    if not known_ident:
                        lg.err('unknown group creator identity: %r' % group_creator_idurl.to_bin())
                        p2p_service.SendFailNoRequest(from_idurl, packet_id, 'unknown group creator identity')
                        continue
                    p2p_service.SendFailNoRequest(from_idurl, packet_id, 'identity:%s' % known_ident.serialize(as_text=True))
                    continue
                lg.warn('skipped incoming queue_message, queue %r is not registered' % queue_id)
                p2p_service.SendFailNoRequest(from_idurl, packet_id, 'queue ID not registered')
                continue
            if not streams()[queue_id]['active']:
                lg.warn('skipped incoming queue_message, queue %r is not active' % queue_id)
                p2p_service.SendFailNoRequest(from_idurl, packet_id, 'queue is not active')
                continue
        if msg_action == 'consume':
            # request from group_participant() to catch up unread messages from the queue
            # TODO: decide about a solution to read past messages and potentially cleanup that block
            consumer_id = msg_data.get('consumer_id')
            if consumer_id not in streams()[queue_id]['consumers']:
                lg.warn('skipped incoming "queue-read" request, consumer %r is not registered for queue %r' % (consumer_id, queue_id))
                p2p_service.SendFailNoRequest(from_idurl, packet_id, 'consumer is not registered')
                continue
            if not streams()[queue_id]['consumers'][consumer_id]['active']:
                lg.warn('skipped incoming "queue-read" request, consumer %r is not active in queue %r' % (consumer_id, queue_id))
                p2p_service.SendFailNoRequest(from_idurl, packet_id, 'consumer is not active')
                continue
            p2p_service.SendAckNoRequest(from_idurl, packet_id)
            handled = True
            continue
        if msg_action == 'produce':
            try:
                payload = msg_data['payload']
                created = msg_data['created']
                producer_id = msg_data['producer_id']
            except:
                lg.exc()
                continue
            if msg_type == 'queue_message':
                if producer_id not in streams()[queue_id]['producers']:
                    lg.warn('skipped incoming queue_message, producer %r is not registered for queue %r' % (producer_id, queue_id))
                    p2p_service.SendFailNoRequest(from_idurl, packet_id, 'producer is not registered')
                    continue
                if not streams()[queue_id]['producers'][producer_id]['active']:
                    lg.warn('skipped incoming queue_message, producer %r is not active in queue %r' % (producer_id, queue_id))
                    p2p_service.SendFailNoRequest(from_idurl, packet_id, 'producer is not active')
                    continue
            # incoming message from group_participant() to push new message to the queue and deliver to all other group members
            received += 1
            if not do_push_message(from_idurl, packet_id, queue_id, producer_id, payload, created):
                continue
            pushed += 1
            handled = True
            continue
        raise Exception('unexpected message "action": %r' % msg_action)
    if received > pushed:
        lg.warn('some of the received messages was not pushed to the queue %r' % queue_id)
    if not handled:
        if _Debug:
            lg.dbg(_DebugLevel, 'queue messages was not handled: %r' % json_messages)
    return handled


def do_push_message(from_idurl, packet_id, queue_id, producer_id, payload, created):
    new_sequence_id = increment_sequence_id(queue_id)
    json_message = make_message(new_sequence_id, producer_id, payload, created)
    if _Debug:
        lg.out(_DebugLevel, '<<< PUSH <<<    into %r by %r at sequence %d' % (queue_id, producer_id, new_sequence_id))
    try:
        p2p_queue.write_message(
            producer_id=producer_id,
            queue_id=queue_id,
            data=json_message,
            creation_time=created,
        )
    except:
        lg.exc()
        return False
    p2p_service.SendAckNoRequest(from_idurl, packet_id)
    return True


#------------------------------------------------------------------------------


def on_message_processed(processed_message):
    sequence_id = processed_message.get_sequence_id()
    if _Debug:
        lg.args(
            _DebugLevel, queue_id=processed_message.queue_id, sequence_id=sequence_id, message_id=processed_message.message_id, success_notifications=processed_message.success_notifications,
            failed_notifications=processed_message.failed_notifications
        )
    if sequence_id is None:
        return False
    if processed_message.failed_notifications:
        if _Debug:
            lg.out(_DebugLevel, '>>> FAILED >>>    from %r at sequence %d, failed_consumers=%d' % (processed_message.queue_id, sequence_id, len(processed_message.failed_notifications)))
    else:
        if _Debug:
            lg.out(_DebugLevel, '>>> PULL >>>    from %r at sequence %d with success count %d' % (processed_message.queue_id, sequence_id, len(processed_message.success_notifications)))
    return True


def on_consumer_notify(message_info):
    try:
        queue_id = message_info['queue_id']
    except:
        lg.exc('invalid incoming message: %r' % message_info)
        return False
    if not queue_id.startswith('group_'):
        # ignore the message, it seems it is not a queue message but it is addressed to the same consumer
        return False
    try:
        payload = message_info['payload']
    except:
        lg.exc('invalid incoming message: %r' % message_info)
        return False
    if 'sequence_id' not in payload:
        # ignore the message, it seems it is not a queue message but it is addressed to the same consumer
        return False
    try:
        consumer_id = message_info['consumer_id']
        packet_id = packetid.MakeQueueMessagePacketID(queue_id, packetid.UniqueID())
        sequence_id = payload['sequence_id']
        last_sequence_id = get_latest_sequence_id(queue_id)
        producer_id = payload['producer_id']
    except:
        lg.exc('invalid incoming message: %r' % message_info)
        return False
    if _Debug:
        lg.args(_DebugLevel, p=producer_id, c=consumer_id, q=queue_id, s=sequence_id, l=last_sequence_id)
    ret = message.send_message(
        json_data={
            'msg_type': 'queue_message',
            'action': 'read',
            'created': utime.utcnow_to_sec1970(),
            'items': [
                {
                    'sequence_id': sequence_id,
                    'created': payload['created'],
                    'producer_id': producer_id,
                    'payload': payload['payload'],
                },
            ],
            'last_sequence_id': last_sequence_id,
        },
        recipient_global_id=my_keys.make_key_id(alias='master', creator_glob_id=consumer_id),
        packet_id=packet_id,
        # message_ack_timeout=config.conf().getInt('services/message-broker/message-ack-timeout'),
        skip_handshake=True,
        fire_callbacks=False,
    )
    ret.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, ignore=True, method='postman.on_consumer_notify')
    if _Debug:
        lg.out(_DebugLevel, '>>> NOTIFY >>>    from %r by producer %r to consumer %r at sequence %d' % (queue_id, producer_id, consumer_id, sequence_id))
    return ret


#------------------------------------------------------------------------------


def get_latest_sequence_id(queue_id):
    if queue_id not in streams():
        return -1
    return streams()[queue_id].get('last_sequence_id', -1)


def increment_sequence_id(queue_id):
    last_sequence_id = int(streams()[queue_id]['last_sequence_id'])
    if last_sequence_id <= 0:
        last_sequence_id = utime.get_milliseconds1970()
    new_last_sequence_id = utime.get_milliseconds1970()
    if last_sequence_id >= new_last_sequence_id:
        new_last_sequence_id = last_sequence_id + 1
    streams()[queue_id]['last_sequence_id'] = new_last_sequence_id
    return new_last_sequence_id


def make_message(sequence_id, producer_id, payload, created):
    return {
        'sequence_id': sequence_id,
        'created': created,
        'producer_id': producer_id,
        'payload': payload,
        'attempts': [],
        'processed': None,
    }


#------------------------------------------------------------------------------


def close_all_streams():
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    list_queues = os.listdir(queues_dir)
    for queue_id in list_queues:
        close_stream(queue_id, erase_data=False)
    return True


def check_rotate_queues():
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    if not os.path.isdir(queues_dir):
        bpio._dirs_make(queues_dir)
    rotated = 0
    known_queueus = os.listdir(queues_dir)
    for queue_id in known_queueus:
        queue_info = global_id.ParseGlobalQueueID(queue_id)
        this_broker_idurl = global_id.glob2idurl(queue_info['supplier_id'])
        if id_url.is_cached(this_broker_idurl) and not this_broker_idurl.is_latest():
            latest_queue_id = global_id.MakeGlobalQueueID(queue_info['queue_alias'], queue_info['owner_id'], this_broker_idurl.to_id())
            if latest_queue_id != queue_id:
                latest_queue_path = os.path.join(queues_dir, latest_queue_id)
                old_queue_path = os.path.join(queues_dir, queue_id)
                if not os.path.isfile(latest_queue_path):
                    bpio.move_dir_recursive(old_queue_path, latest_queue_path)
                    try:
                        bpio._dir_remove(old_queue_path)
                    except:
                        pass
                    rotated += 1
                    lg.info('detected and processed queue rotate : %r -> %r' % (queue_id, latest_queue_id))
                else:
                    bpio._dir_remove(old_queue_path)
                    lg.warn('found an old queue %r and deleted' % old_queue_path)
    if _Debug:
        lg.args(_DebugLevel, rotated=rotated)


#------------------------------------------------------------------------------


def open_stream(queue_id):
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id)
    if queue_id in streams():
        lg.warn('stream already exist: %r' % queue_id)
        return False
    register_stream(queue_id)
    save_stream(queue_id)
    return True


def close_stream(queue_id, erase_data=True):
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id)
    if queue_id not in streams():
        lg.warn('stream not found: %r' % queue_id)
        return False
    if streams()[queue_id]['active']:
        stop_stream(queue_id)
    if erase_data:
        erase_stream(queue_id)
    unregister_stream(queue_id)
    return True


def save_stream(queue_id):
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    queue_dir = os.path.join(queues_dir, queue_id)
    consumers_dir = os.path.join(queue_dir, 'consumers')
    producers_dir = os.path.join(queue_dir, 'producers')
    stream_info = streams()[queue_id]
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, typ=type(queue_id))
    if not os.path.isdir(consumers_dir):
        bpio._dirs_make(consumers_dir)
    if not os.path.isdir(producers_dir):
        bpio._dirs_make(producers_dir)
    for consumer_id, consumer_info in stream_info['consumers'].items():
        local_fs.WriteTextFile(os.path.join(consumers_dir, consumer_id), jsn.dumps(consumer_info))
    for producer_id, producer_info in stream_info['producers'].items():
        local_fs.WriteTextFile(os.path.join(producers_dir, producer_id), jsn.dumps(producer_info))
    return True


def erase_stream(queue_id):
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    queue_dir = os.path.join(queues_dir, queue_id)
    erased_files = 0
    if os.path.isdir(queue_dir):
        erased_files += bpio.rmdir_recursive(queue_dir, ignore_errors=True)
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, queue_dir=queue_dir, erased_files=erased_files)
    return True


def rename_stream(old_queue_id, new_queue_id):
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    old_queue_dir = os.path.join(queues_dir, old_queue_id)
    new_queue_dir = os.path.join(queues_dir, new_queue_id)
    old_customer_idurl = global_id.GetGlobalQueueOwnerIDURL(old_queue_id)
    new_customer_idurl = global_id.GetGlobalQueueOwnerIDURL(new_queue_id)
    if old_customer_idurl not in customers().keys():
        return False
    if old_queue_id not in customers()[old_customer_idurl]:
        return False
    customers()[old_customer_idurl].remove(old_queue_id)
    if not customers()[old_customer_idurl]:
        customers().pop(old_customer_idurl)
    streams()[new_queue_id] = streams().pop(old_queue_id)
    if new_customer_idurl not in customers():
        customers()[new_customer_idurl] = []
    if new_queue_id not in customers()[new_customer_idurl]:
        customers()[new_customer_idurl].append(new_queue_id)
    if os.path.isdir(new_queue_dir):
        bpio.rmdir_recursive(new_queue_dir, ignore_errors=True)
    if os.path.isdir(old_queue_dir):
        bpio.move_dir_recursive(old_queue_dir, new_queue_dir)
    if _Debug:
        lg.args(_DebugLevel, old=old_queue_id, new=new_queue_id)
    return True


#------------------------------------------------------------------------------


def add_consumer(queue_id, consumer_id, consumer_info=None):
    if queue_id not in streams():
        return False
    if consumer_id in streams()[queue_id]['consumers']:
        return False
    if not consumer_info:
        consumer_info = {
            'active': False,
            'last_sequence_id': -1,
        }
    streams()[queue_id]['consumers'][consumer_id] = consumer_info
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, consumer_id=consumer_id, consumer_info=consumer_info)
    if not save_consumer(queue_id, consumer_id):
        raise Exception('failed to save consumer info')
    return True


def remove_consumer(queue_id, consumer_id):
    if queue_id not in streams():
        return False
    if consumer_id not in streams()[queue_id]['consumers']:
        return False
    streams()[queue_id]['consumers'].pop(consumer_id)
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, consumer_id=consumer_id)
    if not erase_consumer(queue_id, consumer_id):
        raise Exception('failed to erase consumer info')
    return True


def save_consumer(queue_id, consumer_id):
    consumer_info = streams()[queue_id]['consumers'][consumer_id]
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    queue_dir = os.path.join(queues_dir, queue_id)
    consumers_dir = os.path.join(queue_dir, 'consumers')
    if not os.path.isdir(consumers_dir):
        bpio._dirs_make(consumers_dir)
    consumer_path = os.path.join(consumers_dir, consumer_id)
    ret = local_fs.WriteTextFile(consumer_path, jsn.dumps(consumer_info))
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, consumer_id=consumer_id, ret=ret)
    return ret


def erase_consumer(queue_id, consumer_id):
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    queue_dir = os.path.join(queues_dir, queue_id)
    consumers_dir = os.path.join(queue_dir, 'consumers')
    consumer_path = os.path.join(consumers_dir, consumer_id)
    if not os.path.isfile(consumer_path):
        return False
    try:
        os.remove(consumer_path)
    except:
        lg.exc()
        return False
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, consumer_id=consumer_id, consumer_path=consumer_path)
    return True


#------------------------------------------------------------------------------


def add_producer(queue_id, producer_id, producer_info=None):
    if queue_id not in streams():
        return False
    if producer_id in streams()[queue_id]['producers']:
        return False
    if not producer_info:
        producer_info = {
            'active': False,
            'last_sequence_id': -1,
        }
    streams()[queue_id]['producers'][producer_id] = producer_info
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, producer_id=producer_id, producer_info=producer_info)
    if not save_producer(queue_id, producer_id):
        raise Exception('failed to store producer info')
    return True


def remove_producer(queue_id, producer_id):
    if queue_id not in streams():
        return False
    if producer_id not in streams()[queue_id]['producers']:
        return False
    streams()[queue_id]['producers'].pop(producer_id)
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, producer_id=producer_id)
    if not erase_producer(queue_id, producer_id):
        raise Exception('failed to erase producer info')
    return True


def save_producer(queue_id, producer_id):
    producer_info = streams()[queue_id]['producers'][producer_id]
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    queue_dir = os.path.join(queues_dir, queue_id)
    producers_dir = os.path.join(queue_dir, 'producers')
    if not os.path.isdir(producers_dir):
        bpio._dirs_make(producers_dir)
    producer_path = os.path.join(producers_dir, producer_id)
    ret = local_fs.WriteTextFile(producer_path, jsn.dumps(producer_info))
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, producer_id=producer_id, ret=ret)
    return ret


def erase_producer(queue_id, producer_id):
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    queue_dir = os.path.join(queues_dir, queue_id)
    producers_dir = os.path.join(queue_dir, 'producers')
    producer_path = os.path.join(producers_dir, producer_id)
    if not os.path.isfile(producer_path):
        return False
    try:
        os.remove(producer_path)
    except:
        lg.exc()
        return False
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, producer_id=producer_id)
    return True


#------------------------------------------------------------------------------


def is_consumer_active(queue_id, consumer_id):
    if queue_id not in streams():
        return False
    if consumer_id not in streams()[queue_id]['consumers']:
        return False
    return streams()[queue_id]['consumers'][consumer_id]['active']


def start_consumer(queue_id, consumer_id):
    if queue_id not in streams():
        lg.warn('queue % is not active, can not start consumer %r' % (queue_id, consumer_id))
        return False
    if consumer_id not in streams()[queue_id]['consumers']:
        lg.warn('not able to start consumer %r because it was not added to the queue %r' % (consumer_id, queue_id))
        return False
    if not p2p_queue.is_consumer_exists(consumer_id):
        p2p_queue.add_consumer(consumer_id)
    if not p2p_queue.is_callback_method_registered(consumer_id, on_consumer_notify):
        p2p_queue.add_callback_method(consumer_id, on_consumer_notify, interested_queues_list=['group_'])
    if not p2p_queue.is_consumer_subscribed(consumer_id, queue_id):
        p2p_queue.subscribe_consumer(consumer_id, queue_id)
    streams()[queue_id]['consumers'][consumer_id]['active'] = True
    save_consumer(queue_id, consumer_id)
    lg.info('consumer %s started in the queue %s' % (consumer_id, queue_id))
    return True


def stop_consumer(queue_id, consumer_id):
    if queue_id not in streams():
        lg.warn('queue % is not active, can not stop consumer %r' % (queue_id, consumer_id))
        return False
    if consumer_id not in streams()[queue_id]['consumers']:
        lg.warn('not able to stop consumer %r because it was not added to the queue %r' % (consumer_id, queue_id))
        return False
    if p2p_queue.is_callback_method_registered(consumer_id, on_consumer_notify):
        p2p_queue.remove_callback_method(consumer_id, on_consumer_notify)
    if p2p_queue.is_consumer_subscribed(consumer_id, queue_id):
        p2p_queue.unsubscribe_consumer(consumer_id, queue_id, remove_empty=True)
    streams()[queue_id]['consumers'][consumer_id]['active'] = False
    save_consumer(queue_id, consumer_id)
    lg.info('consumer %s stopped in the queue %s' % (consumer_id, queue_id))
    return True


#------------------------------------------------------------------------------


def is_producer_active(queue_id, producer_id):
    if queue_id not in streams():
        return False
    if producer_id not in streams()[queue_id]['producers']:
        return False
    return streams()[queue_id]['producers'][producer_id]['active']


def start_producer(queue_id, producer_id):
    if queue_id not in streams():
        lg.warn('queue % is not active, can not start producer %r' % (queue_id, producer_id))
        return False
    if producer_id not in streams()[queue_id]['producers']:
        lg.warn('not able to start producer %r because it was not added to the queue %r' % (producer_id, queue_id))
        return False
    if not p2p_queue.is_producer_exist(producer_id):
        p2p_queue.add_producer(producer_id)
    if not p2p_queue.is_producer_connected(producer_id, queue_id):
        p2p_queue.connect_producer(producer_id, queue_id)
    streams()[queue_id]['producers'][producer_id]['active'] = True
    save_producer(queue_id, producer_id)
    lg.info('producer %s started in the queue %s' % (producer_id, queue_id))
    return True


def stop_producer(queue_id, producer_id):
    if queue_id not in streams():
        lg.warn('queue % is not active, can not stop producer %r' % (queue_id, producer_id))
        return False
    if producer_id not in streams()[queue_id]['producers']:
        lg.warn('not able to stop producer %r because it was not added to the queue %r' % (producer_id, queue_id))
        return False
    if p2p_queue.is_producer_connected(producer_id, queue_id):
        p2p_queue.disconnect_producer(producer_id, queue_id, remove_empty=True)
    streams()[queue_id]['producers'][producer_id]['active'] = False
    save_producer(queue_id, producer_id)
    lg.info('producer %s stopped in the queue %s' % (producer_id, queue_id))
    return True


#------------------------------------------------------------------------------


def is_stream_active(queue_id):
    if queue_id not in streams():
        return False
    return streams()[queue_id]['active']


def start_stream(queue_id):
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id)
    if not p2p_queue.is_queue_exist(queue_id):
        p2p_queue.open_queue(queue_id)
    for consumer_id in list(streams()[queue_id]['consumers'].keys()):
        start_consumer(queue_id, consumer_id)
    for producer_id in list(streams()[queue_id]['producers'].keys()):
        start_producer(queue_id, producer_id)
    streams()[queue_id]['active'] = True
    p2p_queue.touch_queues()
    return True


def stop_stream(queue_id):
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id)
    for producer_id in list(streams()[queue_id]['producers'].keys()):
        stop_producer(queue_id, producer_id)
    for consumer_id in list(streams()[queue_id]['consumers'].keys()):
        stop_consumer(queue_id, consumer_id)
    if p2p_queue.is_queue_exist(queue_id):
        p2p_queue.close_queue(queue_id, remove_empty_consumers=True, remove_empty_producers=True)
    streams()[queue_id]['active'] = False
    p2p_queue.touch_queues()
    return True


def start_all_streams():
    for queue_id, one_stream in streams().items():
        if not one_stream['active']:
            start_stream(queue_id)


def stop_all_streams():
    for queue_id, one_stream in streams().items():
        if one_stream['active']:
            stop_stream(queue_id)


def ping_all_streams():
    target_nodes = list(set(list_customers() + list_consumers_producers()))
    if _Debug:
        lg.args(_DebugLevel, target_nodes=target_nodes)
    propagate.propagate(selected_contacts=target_nodes, wide=True, refresh_cache=True)


#------------------------------------------------------------------------------


def list_customers():
    ret = list(customers().keys())
    if _Debug:
        lg.args(_DebugLevel, r=ret)
    return ret


def list_consumers_producers(include_consumers=True, include_producers=True):
    result = set()
    for queue_id in streams().keys():
        if include_consumers:
            for consumer_id in streams()[queue_id]['consumers']:
                consumer_idurl = global_id.glob2idurl(consumer_id)
                if consumer_idurl not in result:
                    result.add(consumer_idurl)
        if include_producers:
            for producer_id in streams()[queue_id]['producers']:
                producer_idurl = global_id.glob2idurl(producer_id)
                if producer_idurl not in result:
                    result.add(producer_idurl)
    ret = list(result)
    if _Debug:
        lg.args(_DebugLevel, r=ret)
    return ret


#------------------------------------------------------------------------------


def on_queue_connect_request(request_packet, result_defer, consumer_id, producer_id, group_key_info):
    if _Debug:
        lg.args(_DebugLevel, consumer_id=consumer_id, producer_id=producer_id)
    if not my_keys.verify_key_info_signature(group_key_info):
        if _Debug:
            lg.exc('group key verification failed', exc_value=Exception(group_key_info))
        p2p_service.SendFail(request_packet, 'group key verification failed')
        result_defer.callback(False)
        return
    try:
        group_key_id, key_object = my_keys.read_key_info(group_key_info)
    except:
        lg.exc()
        p2p_service.SendFail(request_packet, 'failed reading key info')
        result_defer.callback(False)
        return
    group_key_id = my_keys.latest_key_id(group_key_id)
    group_key_alias, group_creator_idurl = my_keys.split_key_id(group_key_id)
    if not group_key_alias or not group_creator_idurl:
        lg.warn('wrong group_key_id: %r' % group_key_id)
        p2p_service.SendFail(request_packet, 'invalid group_key_id')
        result_defer.callback(False)
        return
    if not group_creator_idurl.is_latest():
        lg.warn('group creator idurl was rotated, consumer must refresh own identity cache: %r ~ %r' % (group_creator_idurl.to_original(), group_creator_idurl.to_bin()))
        known_ident = identitycache.get_one(group_creator_idurl.to_bin())
        if not known_ident:
            lg.err('unknown group creator identity: %r' % group_creator_idurl)
            p2p_service.SendFail(request_packet, 'unknown group creator identity')
            result_defer.callback(False)
            return
        p2p_service.SendFail(request_packet, 'identity:%s' % known_ident.serialize(as_text=True))
        result_defer.callback(False)
        return
    if my_keys.is_key_registered(group_key_id):
        if my_keys.is_key_private(group_key_id):
            p2p_service.SendFail(request_packet, 'private key already registered')
            result_defer.callback(False)
            return
        if my_keys.get_public_key_raw(group_key_id) != key_object.toPublicString():
            my_keys.erase_key(group_key_id)
            if not my_keys.register_key(group_key_id, key_object, group_key_info.get('label', '')):
                p2p_service.SendFail(request_packet, 'key register failed')
                result_defer.callback(False)
                return
    else:
        if not my_keys.register_key(group_key_id, key_object, group_key_info.get('label', '')):
            p2p_service.SendFail(request_packet, 'key register failed')
            result_defer.callback(False)
            return
    consumer_idurl = global_id.glob2idurl(consumer_id)
    producer_idurl = global_id.glob2idurl(producer_id)
    caching_list = []
    if not id_url.is_cached(group_creator_idurl):
        caching_list.append(identitycache.immediatelyCaching(group_creator_idurl))
    if not id_url.is_cached(consumer_idurl):
        caching_list.append(identitycache.immediatelyCaching(consumer_idurl))
    if not id_url.is_cached(producer_idurl):
        caching_list.append(identitycache.immediatelyCaching(producer_idurl))
    dl = DeferredList(caching_list, consumeErrors=True)
    dl.addCallback(lambda _: do_connect_queue(
        request_packet,
        result_defer,
        consumer_id,
        producer_id,
        group_key_info,
        group_creator_idurl,
    ))
    # TODO: need to cleanup registered key in case request was rejected or ID cache failed
    dl.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='postman.on_queue_connect_request')
    dl.addErrback(on_group_creator_idurl_cache_failed, request_packet, result_defer)


def on_queue_disconnect_request(request_packet, result_defer, consumer_id, producer_id, group_key_info, queue_id):
    if not my_keys.verify_key_info_signature(group_key_info):
        if _Debug:
            lg.exc('group key verification failed while stopping consumer or producer', exc_value=Exception(group_key_info))
        p2p_service.SendFail(request_packet, 'group key verification failed while stopping consumer or producer')
        result_defer.callback(False)
        return
    try:
        group_key_id, _ = my_keys.read_key_info(group_key_info)
    except:
        lg.exc()
        p2p_service.SendFail(request_packet, 'failed reading key info')
        result_defer.callback(False)
        return
    group_key_id = my_keys.latest_key_id(group_key_id)
    group_key_alias, group_creator_idurl = my_keys.split_key_id(group_key_id)
    if not group_key_alias or not group_creator_idurl:
        lg.warn('wrong group_key_id: %r' % group_key_id)
        p2p_service.SendFail(request_packet, 'invalid group_key_id')
        result_defer.callback(False)
        return
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id, consumer_id=consumer_id, producer_id=producer_id, request_packet=request_packet)
    if not queue_id or queue_id not in streams():
        lg.warn('queue was not registered already: %r' % queue_id)
        p2p_service.SendAck(request_packet, 'accepted')
        result_defer.callback(True)
        return
    if consumer_id:
        if not stop_consumer(queue_id, consumer_id):
            lg.err('failed to stop consumer %r for the queue %r' % (consumer_id, queue_id))
            p2p_service.SendFail(request_packet, 'failed to stop consumer for the queue')
            result_defer.callback(False)
            return
        if not remove_consumer(queue_id, consumer_id):
            lg.err('failed to remove consumer %r for the queue %r' % (consumer_id, queue_id))
            p2p_service.SendFail(request_packet, 'consumer is not registered for the queue')
            result_defer.callback(False)
            return
    if producer_id:
        if not stop_producer(queue_id, producer_id):
            lg.err('failed to stop producer %r for the queue %r' % (producer_id, queue_id))
            p2p_service.SendFail(request_packet, 'failed to stop producer for the queue')
            result_defer.callback(False)
            return
        if not remove_producer(queue_id, producer_id):
            lg.err('failed to remove producer %r for the queue %r' % (producer_id, queue_id))
            p2p_service.SendFail(request_packet, 'producer is not registered for the queue')
            result_defer.callback(False)
            return
    if not streams()[queue_id]['consumers'] and not streams()[queue_id]['producers']:
        # at least one member must be in the group to keep it alive
        lg.info('no consumers and no producers left, closing queue %r' % queue_id)
        stop_stream(queue_id)
        close_stream(queue_id, erase_data=True)
    customer_idurl = global_id.GetGlobalQueueOwnerIDURL(queue_id)
    if customer_idurl in customers():
        if len(customers()[customer_idurl]) == 0:
            lg.info('no streams left for %r' % customer_idurl)
            # if my_keys.is_key_registered(group_key_id):
            #     lg.info('clean up group key %r' % group_key_id)
            #     my_keys.erase_key(group_key_id)
    p2p_service.SendAck(request_packet, 'accepted')
    result_defer.callback(True)


def on_group_creator_idurl_cache_failed(err, request_packet, result_defer):
    if _Debug:
        lg.args(_DebugLevel, err=err, request_packet=request_packet)
    p2p_service.SendFail(request_packet, 'group creator, consumer or producer idurl cache failed')
    result_defer.callback(False)
    return None


def on_identity_url_changed(evt):
    if _Debug:
        lg.args(_DebugLevel, **evt.data)
    old_idurl = evt.data['old_idurl']
    new_idurl = evt.data['new_idurl']
    if my_id.getIDURL().to_bin() in [new_idurl, old_idurl]:
        lg.warn('my IDURL was rotated, restarting all streams')
        stop_all_streams()
        p2p_queue.remove_message_processed_callback(on_message_processed)
        close_all_streams()
        check_rotate_queues()
        p2p_queue.add_message_processed_callback(on_message_processed)
        start_all_streams()
        return
    new_id = global_id.idurl2glob(new_idurl)
    queues_to_be_closed = []
    if id_url.is_in(old_idurl, customers().keys(), as_field=False):
        for queue_id in customers()[id_url.field(old_idurl)]:
            queues_to_be_closed.append(queue_id)
    do_close_streams(queues_to_be_closed, erase_key=False)
    for queue_id in streams().keys():
        rotated_consumers = []
        rotated_producers = []
        for cur_consumer_id in streams()[queue_id]['consumers']:
            consumer_idurl = global_id.glob2idurl(cur_consumer_id)
            if id_url.is_the_same(consumer_idurl, old_idurl) or id_url.is_the_same(consumer_idurl, new_idurl):
                rotated_consumers.append((cur_consumer_id, new_id))
        for cur_producer_id in streams()[queue_id]['producers']:
            producer_idurl = global_id.glob2idurl(cur_producer_id)
            if id_url.is_the_same(producer_idurl, old_idurl) or id_url.is_the_same(producer_idurl, new_idurl):
                rotated_producers.append((cur_producer_id, new_id))
        for old_consumer_id, new_consumer_id in rotated_consumers:
            old_consumer_info = streams()[queue_id]['consumers'][old_consumer_id]
            if old_consumer_info['active']:
                stop_consumer(queue_id, old_consumer_id)
            remove_consumer(queue_id, old_consumer_id)
            add_consumer(queue_id, new_consumer_id, consumer_info=old_consumer_info)
            if old_consumer_info['active']:
                start_consumer(queue_id, new_consumer_id)
        for old_producer_id, new_producer_id in rotated_producers:
            old_producer_info = streams()[queue_id]['producers'][old_producer_id]
            if old_producer_info['active']:
                stop_producer(queue_id, old_producer_id)
            remove_producer(queue_id, old_producer_id)
            add_producer(queue_id, new_producer_id, producer_info=old_producer_info)
            if old_producer_info['active']:
                start_producer(queue_id, new_producer_id)


#------------------------------------------------------------------------------


def do_connect_queue(request_packet, result_defer, consumer_id, producer_id, group_key_info, target_customer_idurl):
    try:
        target_customer_id = target_customer_idurl.to_id()
        target_queue_alias = group_key_info['alias']
        target_queue_id = global_id.MakeGlobalQueueID(
            queue_alias=target_queue_alias,
            owner_id=target_customer_id,
            supplier_id=my_id.getIDURL().to_id(),
        )
        known_customer_streams = customers().get(target_customer_idurl, [])
        if _Debug:
            lg.args(_DebugLevel, customer=target_customer_idurl, customer_streams=known_customer_streams)
        for current_queue_id in list(p2p_queue.queue().keys()):
            if current_queue_id == target_queue_id:
                continue
            queue_alias, customer_id, _ = global_id.SplitGlobalQueueID(current_queue_id, split_queue_alias=True)
            if target_queue_alias != queue_alias:
                continue
            if target_customer_id != customer_id:
                continue
            p2p_queue.rename_queue(current_queue_id, target_queue_id)
            if current_queue_id not in streams():
                raise Exception('rotated queue %r was not registered' % current_queue_id)
            if target_queue_id in streams():
                lg.warn('rotated queue %r was already registered' % target_queue_id)
            rename_stream(current_queue_id, target_queue_id)
        if target_queue_id not in streams():
            open_stream(target_queue_id)
        if not is_stream_active(target_queue_id):
            start_stream(target_queue_id)
        if consumer_id:
            add_consumer(target_queue_id, consumer_id)
            start_consumer(target_queue_id, consumer_id)
        if producer_id:
            add_producer(target_queue_id, producer_id)
            start_producer(target_queue_id, producer_id)
    except:
        lg.exc()
        p2p_service.SendFail(request_packet, 'failed connecting to the queue')
        result_defer.callback(False)
        return None
    p = {}
    p['queue_id'] = target_queue_id
    p2p_service.SendAck(request_packet, 'accepted:%s' % jsn.dumps(p, keys_to_text=True, values_to_text=True))
    result_defer.callback(True)
    return None


def do_restart_streams():
    loaded_queues = 0
    loaded_consumers = 0
    loaded_producers = 0
    to_be_started = set()
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    if not os.path.isdir(queues_dir):
        bpio._dirs_make(queues_dir)
    known_queues = os.listdir(queues_dir)
    if _Debug:
        lg.args(_DebugLevel, current=len(streams()), known=len(known_queues))
    for queue_id in known_queues:
        queue_info = global_id.ParseGlobalQueueID(queue_id)
        # this_customer_idurl = global_id.glob2idurl(queue_info['owner_id'])
        this_supplier_idurl = global_id.glob2idurl(queue_info['supplier_id'])
        if not this_supplier_idurl.is_latest():
            lg.err('found unclean rotated queue_id: %r' % queue_id)
            continue
        queue_dir = os.path.join(queues_dir, queue_id)
        consumers_dir = os.path.join(queue_dir, 'consumers')
        producers_dir = os.path.join(queue_dir, 'producers')
        if queue_id not in streams():
            customer_idurl = global_id.glob2idurl(queue_info['owner_id'])
            if not id_url.is_cached(customer_idurl):
                lg.err('customer %r IDURL still is not cached, not able to load stream %r' % (customer_idurl, queue_id))
                continue
            try:
                register_stream(queue_id)
            except:
                lg.exc()
                continue
            to_be_started.add(queue_id)
            loaded_queues += 1
        else:
            to_be_started.add(queue_id)
        for consumer_id in (os.listdir(consumers_dir) if os.path.isdir(consumers_dir) else []):
            if consumer_id in streams()[queue_id]['consumers']:
                lg.warn('consumer %r already exist in stream %r' % (consumer_id, queue_id))
                continue
            consumer_info = jsn.loads_text(local_fs.ReadTextFile(os.path.join(consumers_dir, consumer_id)))
            if not consumer_info:
                lg.err('failed reading consumer info %r from %r' % (consumer_id, queue_id))
                continue
            streams()[queue_id]['consumers'][consumer_id] = consumer_info
            streams()[queue_id]['consumers'][consumer_id]['active'] = False
            start_consumer(queue_id, consumer_id)
            loaded_consumers += 1
        for producer_id in (os.listdir(producers_dir) if os.path.isdir(producers_dir) else []):
            if producer_id in streams()[queue_id]['producers']:
                lg.warn('producer %r already exist in stream %r' % (producer_id, queue_id))
                continue
            producer_info = jsn.loads_text(local_fs.ReadTextFile(os.path.join(producers_dir, producer_id)))
            if not producer_info:
                lg.err('failed reading producer info %r from %r' % (producer_id, queue_id))
                continue
            streams()[queue_id]['producers'][producer_id] = producer_info
            streams()[queue_id]['producers'][producer_id]['active'] = False
            loaded_producers += 1
    for queue_id in to_be_started:
        if not is_stream_active(queue_id):
            start_stream(queue_id)
    if _Debug:
        lg.args(_DebugLevel, q=loaded_queues, c=loaded_consumers, p=loaded_producers, s=len(to_be_started))


def do_close_streams(queues_list, erase_key=False):
    if _Debug:
        lg.args(_DebugLevel, queues_list=queues_list)
    for queue_id in queues_list:
        if queue_id not in streams():
            continue
        if streams()[queue_id]['active']:
            for consumer_id in list(streams()[queue_id]['consumers']):
                if consumer_id:
                    if not stop_consumer(queue_id, consumer_id):
                        lg.warn('failed to stop consumer %r in for queue %r' % (consumer_id, queue_id))
                    if not remove_consumer(queue_id, consumer_id):
                        lg.warn('consumer %r is not registered for queue %r' % (consumer_id, queue_id))
            for producer_id in list(streams()[queue_id]['producers']):
                if producer_id:
                    if not stop_producer(queue_id, producer_id):
                        lg.warn('failed to stop producer %r in for queue %r' % (producer_id, queue_id))
                    if not remove_producer(queue_id, producer_id):
                        lg.warn('producer %r is not registered for queue %r' % (producer_id, queue_id))
        stop_stream(queue_id)
        close_stream(queue_id, erase_data=False)
        customer_idurl = global_id.GetGlobalQueueOwnerIDURL(queue_id)
        if customer_idurl in customers():
            if len(customers()[customer_idurl]) == 0:
                lg.info('no streams left for %r' % customer_idurl)
                # group_key_id = global_id.GetGlobalQueueKeyID(queue_id)
                # group_key_id = my_keys.latest_key_id(group_key_id)
                # if erase_key and my_keys.is_key_registered(group_key_id):
                #     lg.info('clean up group key %r' % group_key_id)
                #     my_keys.erase_key(group_key_id)


def do_cache_known_customers():
    service_dir = settings.ServiceDir('service_joint_postman')
    queues_dir = os.path.join(service_dir, 'queues')
    if not os.path.isdir(queues_dir):
        bpio._dirs_make(queues_dir)
    to_be_cached = []
    for queue_id in os.listdir(queues_dir):
        queue_info = global_id.ParseGlobalQueueID(queue_id)
        customer_idurl = global_id.glob2idurl(queue_info['owner_id'])
        if not customer_idurl:
            lg.err('unknown customer IDURL for queue: %r' % queue_id)
            continue
        if not id_url.is_cached(customer_idurl):
            to_be_cached.append(customer_idurl)
    if not to_be_cached:
        return do_restart_streams()
    d = identitycache.start_multiple(to_be_cached)
    d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='postman.do_cache_known_customers')
    d.addBoth(lambda _: do_restart_streams())
    return d



#!/usr/bin/python
# nonblocking.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (nonblocking.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: nonblocking.

This is a wrapper around built-in module ``subprocess.Popen``. Provide
some extended functionality. Can read/write from pipe without blocking
the main thread.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

import os
import subprocess
import platform

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

PIPE = subprocess.PIPE

# Pipe States:
PIPE_EMPTY = 0
PIPE_READY2READ = 1
PIPE_CLOSED = 2

#------------------------------------------------------------------------------

if getattr(subprocess, 'mswindows', None) or platform.uname()[0] == 'Windows':
    from win32file import ReadFile, WriteFile  # @UnresolvedImport
    from win32pipe import PeekNamedPipe  # @UnresolvedImport
    from win32api import TerminateProcess, OpenProcess, CloseHandle  # @UnresolvedImport
    import msvcrt
else:
    import select
    import fcntl  # @UnresolvedImport
    import signal

#------------------------------------------------------------------------------

from bitdust.logs import lg

#------------------------------------------------------------------------------


class Popen(subprocess.Popen):
    """
    This is inherited from ``subprocess.Popen`` class.

    Added some wrappers and platform specific code. Most important
    method added is ``make_nonblocking``.
    """
    err_report = ''

    def __init__(self, args, bufsize=0, executable=None, stdin=None, stdout=None, stderr=None, preexec_fn=None, close_fds=False, shell=False, cwd=None, env=None, universal_newlines=False, startupinfo=None, creationflags=0):

        self.args = args
        subprocess.Popen.__init__(self, args, bufsize, executable, stdin, stdout, stderr, preexec_fn, close_fds, shell, cwd, env, universal_newlines, startupinfo, creationflags)
        if _Debug:
            lg.out(_DebugLevel, 'nonblocking.Popen created')
            lg.out(_DebugLevel, '    stdin=%r' % self.stdin)
            lg.out(_DebugLevel, '    stdout=%r' % self.stdout)
            lg.out(_DebugLevel, '    stderr=%r' % self.stderr)

    def __del__(self):
        try:
            subprocess.Popen.__del__(self)
        except:
            pass
        if _Debug:
            lg.out(_DebugLevel, 'nonblocking.Popen closed')

    def returncode(self):
        return self.returncode

    def recv(self, maxsize=None):
        r = self._recv('stdout', maxsize)
        if r is None:
            return ''
        return r

    def recv_err(self, maxsize=None):
        return self._recv('stderr', maxsize)

    def send_recv(self, input='', maxsize=None):
        return self.send(input), self.recv(maxsize), self.recv_err(maxsize)

    def get_conn_maxsize(self, which, maxsize):
        if maxsize is None:
            maxsize = 1024
        elif maxsize < 1:
            maxsize = 1
        return getattr(self, which), maxsize

    def _close(self, which):
        getattr(self, which).close()
        setattr(self, which, None)

    def state(self):
        return self._state('stdout')

    def make_nonblocking(self):
        """
        Under Linux use built-in method ``fcntl.fcntl`` to make the pipe
        read/write non blocking.
        """
        if getattr(subprocess, 'mswindows', None) or platform.uname()[0] == 'Windows':
            return
        conn, maxsize = self.get_conn_maxsize('stdout', None)
        if conn is None:
            return
        flags = fcntl.fcntl(conn, fcntl.F_GETFL)
        if not conn.closed:
            fcntl.fcntl(conn, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    if getattr(subprocess, 'mswindows', None) or platform.uname()[0] == 'Windows':

        def send(self, input):
            if not self.stdin:
                return None

            try:
                x = msvcrt.get_osfhandle(self.stdin.fileno())  # @UndefinedVariable
                (errCode, written) = WriteFile(x, input)
            except:
                return None

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None

            try:
                x = msvcrt.get_osfhandle(conn.fileno())  # @UndefinedVariable
                (read, nAvail, nMessage) = PeekNamedPipe(x, 0)
                if maxsize < nAvail:
                    nAvail = maxsize
                if nAvail > 0:
                    (errCode, read) = ReadFile(x, nAvail, None)
            except:
                return None

            if self.universal_newlines:
                read = self._translate_newlines(read)

            return read

        def _state(self, which):
            conn, maxsize = self.get_conn_maxsize(which, None)
            if conn is None:
                return PIPE_CLOSED
            try:
                x = msvcrt.get_osfhandle(conn.fileno())  # @UndefinedVariable
            except:
                return PIPE_CLOSED
            try:
                (read, nAvail, nMessage) = PeekNamedPipe(x, 0)
            except:
                return PIPE_CLOSED
            if nAvail > 0:
                return PIPE_READY2READ
            return PIPE_EMPTY

        def kill(self):
            try:
                PROCESS_TERMINATE = 1
                handle = OpenProcess(PROCESS_TERMINATE, False, self.pid)
                TerminateProcess(handle, -1)
                CloseHandle(handle)
            except:
                pass

    else:

        def send(self, input):
            if not self.stdin:
                return None

            if not select.select([], [self.stdin], [], 0)[1]:
                return None

            try:
                written = os.write(self.stdin.fileno(), input)
            except:
                return None

            return written

        def _recv(self, which, maxsize):
            conn, maxsize = self.get_conn_maxsize(which, maxsize)
            if conn is None:
                return None

            flags = fcntl.fcntl(conn, fcntl.F_GETFL)
            if not conn.closed:
                fcntl.fcntl(conn, fcntl.F_SETFL, flags | os.O_NONBLOCK)

            try:
                if not select.select([conn], [], [], 0)[0]:
                    return None

                r = conn.read(maxsize)
                if not r:
                    return None

                if self.universal_newlines:
                    r = self._translate_newlines(r)
                return r
            finally:
                if not conn.closed:
                    fcntl.fcntl(conn, fcntl.F_SETFL, flags)

        def _state(self, which):
            conn, maxsize = self.get_conn_maxsize(which, None)
            if conn is None:
                return PIPE_CLOSED

            try:
                # check and see if there is any input ready
                ready = select.select([conn], [], [], 0)
                if conn in ready[0]:
                    return PIPE_READY2READ
                return PIPE_EMPTY
            except:
                return PIPE_CLOSED

        def kill(self):
            os.kill(self.pid, signal.SIGTERM)


def ExecuteString(execstr):
    """
    An old method.
    """
    try:
        import win32process  # @UnresolvedImport
        return Popen(
            execstr,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=win32process.CREATE_NO_WINDOW,
        )
    except:
        return None



#!/usr/bin/python
# terminal_chat.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (terminal_chat.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: terminal_chat
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import time
import random
import sys
import termios
import tty
import threading

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.lib import nameurl

from bitdust.userid import global_id

from bitdust.chat import kbhit

#------------------------------------------------------------------------------

_SimpleTerminalChat = None

#------------------------------------------------------------------------------


def init(do_send_message_func=None, do_search_user_func=None):
    """

    """
    global _SimpleTerminalChat
    _SimpleTerminalChat = SimpleTerminalChat(
        send_message_func=do_send_message_func,
        search_user_func=do_search_user_func,
    )


def shutdown():
    """

    """
    global _SimpleTerminalChat
    del _SimpleTerminalChat
    _SimpleTerminalChat = None


def run():
    global _SimpleTerminalChat
    _SimpleTerminalChat.run()
    reactor.callFromThread(reactor.stop)


def stop():
    global _SimpleTerminalChat
    _SimpleTerminalChat.stop()


def process_message(sender, message):
    global _SimpleTerminalChat
    _SimpleTerminalChat.process_message(sender, message)


def on_incoming_message(msg):
    global _SimpleTerminalChat
    _SimpleTerminalChat.on_inbox_message(msg['sender'], msg['message'])
    return msg


#------------------------------------------------------------------------------


class SimpleTerminalChat(object):
    def __init__(self, send_message_func=None, search_user_func=None):
        self.chars = []
        self.input = []
        self.history = []
        self.printed = 0
        self.quitnow = 0
        self.users = []
        self.send_message_func = send_message_func
        self.search_user_func = search_user_func

    def on_inbox_message(self, sender, message):
        name = nameurl.GetName(sender)
        if sender not in self.users:
            self.users.append(sender)
            self.history.append({
                'text': 'user %s was joined' % name,
                'name': '',
                'time': time.time(),
            })
        self.history.append({
            'text': message,
            'name': nameurl.GetName(sender),
            'sender': sender,
            'time': time.time(),
        })

    def on_nickname_search_result(self, results):
        if results['status'] != 'OK':
            self.history.append({
                'text': 'search failed: %s' % results['errors'],
                'name': '',
                'time': time.time(),
            })
            return None
        for r in results['result']:
            self.history.append({
                'text': '%s' % (r['idurl'] if r['idurl'] else 'not found'),
                'name': '',
                'time': time.time(),
            })

    def on_my_message(self, message):
        if message.startswith('!add '):
            idurl = message[5:]
            if global_id.IsValidGlobalUser(idurl):
                gid = global_id.NormalizeGlobalID(idurl)
                idurl = gid['idurl']
            if idurl.strip() and idurl not in self.users:
                self.users.append(idurl)
                name = nameurl.GetName(idurl)
                self.history.append({
                    'text': 'user "%s" was added to the channel' % name,
                    'name': '',
                    'time': time.time(),
                })
            return
        if message.startswith('!find ') or message.startswith('!search '):
            _, _, inp = message.partition(' ')
            if not self.search_user_func:
                self.history.append({
                    'text': 'search failed, method not defined',
                    'name': '',
                    'time': time.time(),
                })
                return
            self.search_user_func(inp).addBoth(self.on_nickname_search_result)
            self.history.append({
                'text': 'looking for "%s" ...' % inp,
                'name': '',
                'time': time.time(),
            })
            return
        self.history.append({
            'text': message,
            'name': 'you',
            'time': time.time(),
        })
        if self.send_message_func is not None:
            for to in self.users:
                reactor.callFromThread(self.send_message_func, to, message)

    def bot(self):
        while True:
            if self.quitnow:
                break
            time.sleep(0.1)
            if random.randint(1, 500) == 1:
                self.on_inbox_message(
                    'http://p2p-id.ru/bot.xml',
                    'HI man!    ' + time.asctime(),
                )

    def collect_output(self):
        last_line = ''
        sys.stdout.write('> ')
        sys.stdout.flush()
        while True:
            if self.quitnow:
                break
            time.sleep(0.05)
            if self.printed < len(self.history):
                sys.stdout.write('\b'*(len(last_line) + 2))
                # sys.stdout.write('\n\r')
                sys.stdout.flush()
                for h in self.history[self.printed:]:
                    out = ''
                    if h.get('time'):
                        out += '[%s] ' % time.strftime('%H:%M:%S', time.gmtime(h['time']))
                    if h.get('name'):
                        out += h['name'] + ': '
                    out += h.get('text', '')
                    last_line = out
                    sys.stdout.write(out + '\n\r')
                    sys.stdout.flush()
                sys.stdout.write('> ' + (''.join(self.chars)))
                sys.stdout.flush()
                self.printed = len(self.history)

    def collect_input(self):
        try:
            while True:
                if self.quitnow:
                    break
                ch = self.kb.getch()
                self.input.append(ch)
        except KeyboardInterrupt:
            self.quitnow = True
        return None

    def process_input(self):
        while True:
            if self.quitnow:
                break
            if len(self.input) == 0:
                time.sleep(0.05)
                continue
            inp = list(self.input)
            self.input = []
            # COMBINATION OR SPECIAL KEY PRESSED
            if len(inp) == 3 and ord(inp[0]) == 27:
                continue
            for c in inp:
                # ESC
                if ord(c) == 27:
                    sys.stdout.write('\n\r')
                    sys.stdout.flush()
                    self.quitnow = True
                    break
                # BACKSPACE
                if c == '\x7f':
                    if self.chars:
                        del self.chars[-1]
                        sys.stdout.write('\b \b')
                        sys.stdout.flush()
                        continue
                # ENTER
                if c in '\n\r':
                    if len(self.chars) == 0:
                        continue
                    msg = ''.join(self.chars)
                    self.chars = []
                    if msg.strip() in [
                        '!q',
                        '!quit',
                        '!exit',
                    ]:
                        sys.stdout.write('\n\r')
                        sys.stdout.flush()
                        self.quitnow = True
                        return
                    sys.stdout.write('\b'*(len(msg)))
                    sys.stdout.flush()
                    self.on_my_message(msg)
                    continue
                # some printable char
                if ord(c) >= 32 and ord(c) <= 126:
                    sys.stdout.write(c)
                    sys.stdout.flush()
                    self.chars.append(c)

    def welcome(self):
        sys.stdout.write('type your message and press Enter to send on channel\n\r')
        sys.stdout.write('use "!add <idurl>" command to invite people here\n\r')
        sys.stdout.write('press ESC or send "!q" to quit\n\r')
        sys.stdout.flush()

    def goodbye(self):
        sys.stdout.write('Good luck to you!\n\r')
        sys.stdout.flush()

    def run(self):
        self.fd = sys.stdin.fileno()
        self.old_settings = termios.tcgetattr(self.fd)
        self.kb = kbhit.KBHit()
        tty.setraw(sys.stdin.fileno())

        self.welcome()
        # bot = threading.Thread(target=self.bot)
        # bot.start()
        out = threading.Thread(target=self.collect_output)
        out.start()
        inp = threading.Thread(target=self.collect_input)
        inp.start()
        proc = threading.Thread(target=self.process_input)
        proc.start()
        proc.join()
        self.goodbye()

        termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)
        self.kb.set_normal_term()

    def stop(self):
        self.quitnow = 1


#------------------------------------------------------------------------------

if __name__ == '__main__':
    init()
    reactor.callInThread(run)
    reactor.run()
    shutdown()



#!/usr/bin/python
# service_p2p_hookups.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_p2p_hookups.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_p2p_hookups
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return P2PHookupsService()


class P2PHookupsService(LocalService):

    service_name = 'service_p2p_hookups'
    config_path = 'services/p2p-hookups/enabled'

    def dependent_on(self):
        from bitdust.main import settings
        depends = [
            'service_gateway',
            'service_identity_propagate',
        ]
        if settings.enableTCP():
            depends.append('service_tcp_transport')
        if settings.enableUDP():
            depends.append('service_udp_transport')
        return depends

    def start(self):
        from twisted.internet.defer import Deferred
        from bitdust.transport import callback
        from bitdust.main import events
        from bitdust.main import listeners
        from bitdust.p2p import online_status
        from bitdust.p2p import p2p_service
        from bitdust.p2p import p2p_connector
        from bitdust.p2p import network_connector
        from bitdust.p2p import ratings
        p2p_service.init()
        online_status.init()
        ratings.init()
        self._starting_defer = Deferred()
        p2p_connector.A('init')
        p2p_connector.A().addStateChangedCallback(self._on_p2p_connector_switched)
        network_connector.A().addStateChangedCallback(self._on_network_connector_switched)
        callback.append_inbox_callback(self._on_inbox_packet_received)
        callback.append_inbox_callback(p2p_service.inbox)
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        events.add_subscriber(self._on_my_identity_url_changed, 'my-identity-url-changed')
        if listeners.is_populate_required('online_status'):
            # listeners.populate_later().remove('online_status')
            online_status.populate_online_statuses()
        return True

    def stop(self):
        from bitdust.transport import callback
        from bitdust.main import events
        from bitdust.p2p import online_status
        from bitdust.p2p import p2p_service
        from bitdust.p2p import p2p_connector
        from bitdust.p2p import network_connector
        from bitdust.p2p import ratings
        events.remove_subscriber(self._on_my_identity_url_changed, 'my-identity-url-changed')
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        callback.remove_inbox_callback(p2p_service.inbox)
        if network_connector.A():
            network_connector.A().removeStateChangedCallback(self._on_network_connector_switched)
        p2p_connector.A().removeStateChangedCallback(self._on_p2p_connector_switched)
        ratings.shutdown()
        online_status.shutdown()
        p2p_connector.Destroy()
        p2p_service.shutdown()
        return True

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        from bitdust.p2p import commands
        if newpacket.Command == commands.RequestService():
            return self._on_request_service_received(newpacket, info)
        elif newpacket.Command == commands.CancelService():
            return self._on_cancel_service_received(newpacket, info)
        return False

    def _on_request_service_received(self, newpacket, info):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.lib import serialization
        from bitdust.services import driver
        from bitdust.p2p import p2p_service
        from bitdust.transport import packet_out
        if len(newpacket.Payload) > 1024*10:
            lg.warn('too long payload')
            p2p_service.SendFail(newpacket, 'too long payload')
            return True
        try:
            json_payload = serialization.BytesToDict(newpacket.Payload, keys_to_text=True, values_to_text=True)
            json_payload['name']
            json_payload['payload']
        except:
            lg.warn('json payload invalid')
            p2p_service.SendFail(newpacket, 'json payload invalid')
            return True
        service_name = str(json_payload['name'])
        lg.out(self.debug_level, 'service_p2p_hookups.RequestService {%s} from %s' % (service_name, newpacket.OwnerID))
        if not driver.is_exist(service_name):
            lg.warn('got wrong payload in %s' % service_name)
            p2p_service.SendFail(newpacket, 'service %s not exist' % service_name)
            return True
        if not driver.is_on(service_name):
            p2p_service.SendFail(newpacket, 'service %s is off' % service_name)
            return True
        try:
            result = driver.request(service_name, json_payload['payload'], newpacket, info)
        except:
            lg.exc()
            p2p_service.SendFail(newpacket, 'request processing failed with exception')
            return True
        if not result:
            lg.out(self.debug_level, 'service_p2p_hookups._send_request_service SKIP request %s' % service_name)
            return False
        if isinstance(result, Deferred):
            lg.out(self.debug_level, 'service_p2p_hookups._send_request_service fired delayed execution')
        elif isinstance(result, packet_out.PacketOut):
            lg.out(self.debug_level, 'service_p2p_hookups._send_request_service outbox packet sent')
        return True

    def _on_cancel_service_received(self, newpacket, info):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.lib import serialization
        from bitdust.services import driver
        from bitdust.p2p import p2p_service
        from bitdust.transport import packet_out
        if len(newpacket.Payload) > 1024*10:
            p2p_service.SendFail(newpacket, 'too long payload')
            return True
        try:
            json_payload = serialization.BytesToDict(newpacket.Payload, keys_to_text=True, values_to_text=True)
            json_payload['name']
            json_payload['payload']
        except:
            p2p_service.SendFail(newpacket, 'json payload invalid')
            return True
        service_name = json_payload['name']
        lg.out(self.debug_level, 'service_p2p_hookups.CancelService {%s} from %s' % (service_name, newpacket.OwnerID))
        if not driver.is_exist(service_name):
            lg.warn('got wrong payload in %s' % newpacket)
            p2p_service.SendFail(newpacket, 'service %s not exist' % service_name)
            return True
        if not driver.is_on(service_name):
            p2p_service.SendFail(newpacket, 'service %s is off' % service_name)
            return True
        try:
            result = driver.cancel(service_name, json_payload['payload'], newpacket, info)
        except:
            lg.exc()
            p2p_service.SendFail(newpacket, 'request processing failed with exception')
            return True
        if not result:
            lg.out(self.debug_level, 'service_p2p_hookups._send_cancel_service SKIP request %s' % service_name)
            return False
        if isinstance(result, Deferred):
            lg.out(self.debug_level, 'service_p2p_hookups._send_cancel_service fired delayed execution')
        elif isinstance(result, packet_out.PacketOut):
            lg.out(self.debug_level, 'service_p2p_hookups._send_cancel_service outbox packet sent')
        return True

    def _on_p2p_connector_switched(self, oldstate, newstate, evt, *args, **kwargs):
        if newstate == 'INCOMMING?':
            if self._starting_defer is not None:
                self._starting_defer.callback(newstate)
                self._starting_defer = None

    def _on_network_connector_switched(self, oldstate, newstate, evt, *args, **kwargs):
        from bitdust.p2p import p2p_connector
        if oldstate != newstate:
            if newstate == 'CONNECTED' or newstate == 'DISCONNECTED':
                p2p_connector.A('network_connector.state', newstate)

    def _on_identity_url_changed(self, evt):
        from twisted.internet import reactor  # @UnresolvedImport
        from bitdust.logs import lg
        from bitdust.userid import id_url
        from bitdust.userid import global_id
        from bitdust.p2p import online_status
        for idurl, inst in online_status.online_statuses().items():
            if idurl == id_url.field(evt.data['old_idurl']):
                idurl.refresh(replace_original=True)
                inst.idurl.refresh(replace_original=True)
                inst.name = 'online_%s' % global_id.UrlToGlobalID(idurl)
                inst.automat('shook-up-hands')
                reactor.callLater(0, inst.automat, 'ping-now')  # @UndefinedVariable
                lg.info('found %r with rotated identity and refreshed: %r' % (inst, idurl))

    def _on_my_identity_url_changed(self, evt):
        from bitdust.services import driver
        if driver.is_on('service_entangled_dht'):
            from bitdust.dht import dht_service
            from bitdust.userid import my_id
            if my_id.getIDURL():
                dht_service.set_node_data('idurl', my_id.getIDURL().to_text())



#!/usr/bin/env python
# mine.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (mine.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import time
import random
import string
import hashlib

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.userid import my_id

from bitdust.lib import utime

from bitdust.coins import coins_io

#------------------------------------------------------------------------------


def init():
    pass


def shutdown():
    pass


#------------------------------------------------------------------------------


def build_starter(length=5):
    return (''.join([random.choice(string.uppercase + string.lowercase + string.digits) for _ in range(length)])) + '_'  # @UndefinedVariable


def build_hash(payload):
    return hashlib.sha1(payload).hexdigest()


def get_hash_complexity(hexdigest, simplification):
    complexity = 0
    while complexity < len(hexdigest):
        if int(hexdigest[complexity], 16) < simplification:
            complexity += 1
        else:
            break
    return complexity


def get_hash_difficulty(hexdigest, simplification=2):
    difficulty = 0
    while True:
        ok = False
        for simpl in range(simplification):
            if hexdigest.startswith(str(simpl)*difficulty):
                ok = True
                break
        if ok:
            difficulty += 1
        else:
            break
    return difficulty - 1


def work_with_known_difficulty(coin_json, difficulty, simplification=2, starter_length=10, starter_limit=99999, stop_marker=None, prev_hash=''):
    coin_json['miner'] = {
        'idurl': my_id.getIDURL().to_bin(),
    }
    # data_dump = json.dumps(data)
    data_dump = coins_io.coin_to_string(coin_json)
    starter = build_starter(starter_length)
    on = 0
    while True:
        if stop_marker is not None and callable(stop_marker):
            if stop_marker():
                return None
        check = starter + str(on)
        # if coin_json is not None:
        #     check += data_dump
        hexdigest = build_hash(check)
        if difficulty == get_hash_complexity(hexdigest, simplification):
            break
        on += 1
        if on > starter_limit:
            starter = build_starter(starter_length)
            on = 0
    coin_json['miner'].update({
        'hash': hexdigest,
        'prev': prev_hash,
        'starter': starter + str(on),
        'mined': utime.utcnow_to_sec1970(),
    })
    return coin_json


def work_from_known_hash(coin_json, prev_hash, simplification=2, starter_length=10, starter_limit=99999, stop_marker=None):
    # data.update({'prev': prev_hash, })
    difficulty = get_hash_difficulty(prev_hash)
    complexity = get_hash_complexity(prev_hash, simplification)
    if difficulty == complexity:
        complexity += 1
        # print 'found golden coin, step up difficulty:', difficulty
    return work_with_known_difficulty(
        coin_json,
        complexity,
        starter_limit=starter_limit,
        starter_length=starter_length,
        simplification=simplification,
        stop_marker=stop_marker,
        prev_hash=prev_hash,
    )


#------------------------------------------------------------------------------


class MininigCounter(object):
    def __init__(self, max_counts, max_seconds):
        self.max_counts = max_counts
        self.max_seconds = max_seconds
        self.counts = 0
        self.started = int(time.time())

    def marker(self):
        if self.counts >= self.max_counts:
            return True
        if int(time.time()) - self.started > self.max_seconds:
            return True
        self.counts += 1
        return False


def _test():
    try:
        for test_no in range(1, 21):
            print('start test N', test_no)
            start = time.time()
            mc = MininigCounter(10**7, 300)
            simplification = 2
            starter_length = 5
            starter_limit = 99999
            coins = 0
            hexhash = ''
            data = {
                'a': 'b',
            }
            while True:
                coin = work_from_known_hash(data, hexhash, simplification=simplification, starter_length=starter_length, starter_limit=starter_limit, stop_marker=mc.marker)
                if not coin:
                    break
                coins += 1
                hexhash = coin['hash']
                print(coin)

            print(coins, 'coins')
            print('last hash:', hexhash)
            print('complexity riched:', get_hash_complexity(hexhash, simplification))
            print(time.time() - start, 'seconds')
            print(mc.counts, 'iterations')
            print('simplification:', simplification)
            print()

    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    init()
    _test()
    shutdown()



#!/usr/bin/env python
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (tar_file.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: tar_file.

This python code can be used to replace the Unix tar command
and so be portable to non-unix machines.
There are other python tar libraries, but this is included with Python.
So that file is starter as child process of BitDust to prepare data for backup.

.. warning:: Note that we should not print things here because tar output goes to standard out.
             If we print anything else to stdout the .tar file will be ruined.
             We must also not print things in anything this calls.

Inspired from examples here:

* `http://docs.python.org/lib/tar-examples.html`
* `http://code.activestate.com/recipes/299412`

This is also working without creating a subprocess:
you can use fileobj input parameter in writetar() method.

TODO:
If we kept track of how far we were through a list of files, and broke off
new blocks at file boundaries, we could restart a backup and continue
were we left off if a crash happened while we were waiting to send a block
(most of the time is waiting so good chance).
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import six
from io import open

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import os
import sys
import platform
import tarfile
import traceback

#------------------------------------------------------------------------------

AppData = ''
_ExcludeFunction = None

#------------------------------------------------------------------------------

if sys.version_info[0] == 3:
    text_type = str
    binary_type = bytes
else:
    text_type = unicode  # @UndefinedVariable
    binary_type = str

#------------------------------------------------------------------------------


def is_text(s):
    """
    Return `True` if `s` is a text value:
    + `unicode()` in Python2
    + `str()` in Python3
    """
    return isinstance(s, text_type)


def is_bin(s):
    """
    Return `True` if `s` is a binary value:
    + `str()` in Python2
    + `bytes()` in Python3
    """
    return isinstance(s, binary_type)


def is_string(s):
    """
    Return `True` if `s` is text or binary type (not integer, class, list, etc...)
    """
    return is_text(s) or is_bin(s)


def to_text(s, encoding='utf-8', errors='strict'):
    """
    If ``s`` is binary type - decode it to unicode - "text" type in Python3 terms.
    If ``s`` is not binary and not text calls `str(s)` to build text representation.
    """
    if s is None:
        return s
    if not is_string(s):
        s = text_type(s)
    if is_text(s):
        return s
    return s.decode(encoding=encoding, errors=errors)


def sharedPath(filename, subdir='logs'):
    global AppData
    if AppData == '':
        curdir = os.getcwd()  # os.path.dirname(os.path.abspath(sys.executable))
        if os.path.isfile(os.path.join(curdir, 'appdata')):
            try:
                appdata = os.path.abspath(open(os.path.join(curdir, 'appdata'), 'rb').read().strip())
            except:
                appdata = os.path.join(os.path.expanduser('~'), '.bitdust')
            if not os.path.isdir(appdata):
                appdata = os.path.join(os.path.expanduser('~'), '.bitdust')
        else:
            if sys.executable == 'android_python' or ('ANDROID_ARGUMENT' in os.environ or 'ANDROID_ROOT' in os.environ):
                from android.storage import app_storage_path  # @UnresolvedImport
                appdata = os.path.join(app_storage_path(), '.bitdust')
            else:
                appdata = os.path.join(os.path.expanduser('~'), '.bitdust')
        AppData = appdata
    return os.path.join(AppData, subdir, filename)


def logfilepath():
    """
    A method to detect where is placed the log file for ``tar_file`` child
    process.
    """
    return sharedPath('tar_file.log')


def printlog(txt, mode='a'):
    """
    Write a line to the log file.
    """
    # try:
    path = logfilepath()
    if not os.path.isfile(path):
        try:
            os.makedirs(os.path.dirname(path))
        except:
            pass
    LogFile = open(path, mode)
    LogFile.write(to_text(txt))
    LogFile.close()
    # except:
    #     pass


def printexc():
    """
    Write exception info to the log file.
    """
    if _Debug:
        printlog('\n' + traceback.format_exc() + '\n')


#------------------------------------------------------------------------------


def LinuxExcludeFunction(source_path, tar_path):
    """
    Return True if given file must not be included in the backup. Filename
    comes in with the path relative to the start path, so:
    "dirbeingbackedup/photos/christmas2008.jpg".

    PREPRO:
    On linux we should test for the attribute meaning "nodump" or "nobackup"
    This is set with:
        chattr +d <file>
    And listed with:
        lsattr <file>
    Also should test that the file is readable and maybe that directory is executable.
    If tar gets stuff it can not read - it just stops and we the whole process is failed.
    """
    # TODO: - must do more smart checking
    # if filename.count(".bitdust"):
    #     return True
    if not os.access(source_path, os.R_OK):
        return True
    return False  # don't exclude the file


def WindowsExcludeFunction(source_path, tar_path):
    """
    Same method for Windows platforms. Filename comes in with the path relative
    to the start path, so: "Local Settings/Application Data/Microsoft/Windows/UsrClass.dat"

    PREPRO: On windows I run into some files that Windows tells me I
    don't have permission to open (system files), I had hoped to use
    os.access(filename, os.R_OK) == False to skip a file if I couldn't
    read it, but I did not get it to work every time. DWC.
    """
    # TODO: - must do more smart checking
    # if source_path.lower().count(".bitdust"):
    #     return True
    if (source_path.lower().find('local settings\\temp') != -1):
        return True
    if not os.access(source_path, os.R_OK):
        return True
    return False  # don't exclude the file


#------------------------------------------------------------------------------

_ExcludeFunction = LinuxExcludeFunction
if platform.uname()[0] == 'Windows':
    _ExcludeFunction = WindowsExcludeFunction

#------------------------------------------------------------------------------


def writetar_filter(tarinfo, sourcepath):
    global _ExcludeFunction
    if _ExcludeFunction(sourcepath, tarinfo.name):
        return None
    return tarinfo


#------------------------------------------------------------------------------


def writetar(sourcepath, arcname=None, subdirs=True, compression='none', encoding=None, fileobj=None, mode=None):
    """
    Create a tar archive from given ``sourcepath`` location.
    """
    global _ExcludeFunction
    if _Debug:
        printlog('WRITE: %s arcname=%s, subdirs=%s, compression=%s, encoding=%s\n' % (sourcepath, arcname, subdirs, compression, encoding))
    if not fileobj:
        fileobj = sys.stdout
    if not mode:
        if compression != 'none':
            mode = 'w|' + compression
        else:
            mode = 'w|tar'
    _, filename = os.path.split(sourcepath)
    if arcname is None:
        arcname = to_text(filename)
    else:
        arcname = to_text(arcname)
    # DEBUG: tar = tarfile.open('', mode, fileobj=open('out.tar', 'wb'), encoding=encoding)
    if _Debug:
        printlog('OPEN: mode=%s fileobj=%r\n' % (mode, fileobj))
    tar = tarfile.open('', mode, fileobj=fileobj, encoding=encoding, bufsize=1024*1024)
    if _Debug:
        printlog('ADD: name=%s arcname=%r\n' % (sourcepath, arcname))
    tar.add(
        name=sourcepath,
        arcname=arcname,
        recursive=subdirs,
        filter=lambda tarinfo: writetar_filter(tarinfo, sourcepath),
    )
    if not subdirs and os.path.isdir(sourcepath):
        # the True is for recursive, if we wanted to just do the immediate directory set to False
        for subfile in os.listdir(sourcepath):
            subpath = os.path.join(sourcepath, subfile)
            if not os.path.isdir(subpath):
                arcname = to_text(os.path.join(arcname, subfile))
                if _Debug:
                    printlog('ADD: name=%s arcname=%r\n' % (subpath, arcname))
                tar.add(
                    name=subpath,
                    arcname=arcname,
                    recursive=False,
                    filter=lambda tarinfo: writetar_filter(tarinfo, subpath),
                )
    tar.close()
    return True


#------------------------------------------------------------------------------


def readtar(archivepath, outputdir, encoding=None, mode='r:*'):
    """
    Extract tar file from ``archivepath`` location into local ``outputdir``
    folder.
    """
    if _Debug:
        printlog('READ: mode=%s name=%s outputdir=%s encoding=%s\n' % (mode, archivepath, outputdir, encoding))
    tar = tarfile.open(name=archivepath, mode=mode, encoding=encoding)
    tar.extractall(outputdir)
    tar.close()
    return True


#------------------------------------------------------------------------------


def main():
    """
    The entry point of the ``tar_file`` child process.

    Use command line arguments to get the command from ``bpmain``.
    """
    ostype = platform.uname()[0]

    if ostype == 'Windows':
        if six.PY3:
            sys.stdout = sys.stdout.buffer
        else:
            try:
                import msvcrt
                msvcrt.setmode(1, os.O_BINARY)  # @UndefinedVariable
            except:
                pass

    else:
        if six.PY3:
            sys.stdout = sys.stdout.buffer

    if len(sys.argv) < 4:
        if _Debug:
            printlog('tar_file.py extract <archive path> <output dir>\n')
            printlog('tar_file.py <subdirs / nosubdirs> <"none" / "bz2" / "gz"> <folder/file path> [archive filename]\n')
        return 2

    try:
        cmd = sys.argv[1].strip().lower()

        if cmd == 'extract':
            readtar(
                archivepath=sys.argv[2],
                outputdir=sys.argv[3],
                encoding='utf-8',
            )

        else:
            arcname = None
            if len(sys.argv) >= 5:
                arcname = sys.argv[4]
            writetar(
                sourcepath=sys.argv[3],
                arcname=arcname,
                subdirs=True if cmd == 'subdirs' else False,
                compression=sys.argv[2],
                encoding='utf-8',
            )
    except:
        printexc()
        return 1

    return 0


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/python
# service_proxy_server.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_proxy_server.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_proxy_server
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return ProxyServerService()


class ProxyServerService(LocalService):

    service_name = 'service_proxy_server'
    config_path = 'services/proxy-server/enabled'

    def dependent_on(self):
        return [
            'service_p2p_hookups',
        ]

    def enabled(self):
        from bitdust.main import settings
        if settings.transportIsEnabled('proxy'):
            return False
        return settings.enableProxyServer()

    def attached_dht_layers(self):
        from bitdust.dht import dht_records
        return [
            dht_records.LAYER_PROXY_ROUTERS,
        ]

    def start(self):
        from bitdust.logs import lg
        from bitdust.services import driver
        from bitdust.main import events
        from bitdust.transport.proxy import proxy_router
        proxy_router.A('init')
        proxy_router.A('start')
        if driver.is_on('service_entangled_dht'):
            self._do_connect_proxy_routers_dht_layer()
        else:
            lg.warn('service service_entangled_dht is OFF')
        events.add_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        return True

    def stop(self):
        from bitdust.services import driver
        from bitdust.main import events
        from bitdust.transport.proxy import proxy_router
        events.remove_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        proxy_router.A('stop')
        proxy_router.A('shutdown')
        if driver.is_on('service_entangled_dht'):
            from bitdust.dht import dht_service
            from bitdust.dht import dht_records
            dht_service.suspend(layer_id=dht_records.LAYER_PROXY_ROUTERS)
        return True

    def request(self, json_payload, newpacket, info):
        from bitdust.transport.proxy import proxy_router
        proxy_router.A('request-route-received', (
            json_payload,
            newpacket,
            info,
        ))
        return True

    def cancel(self, json_payload, newpacket, info):
        from bitdust.transport.proxy import proxy_router
        proxy_router.A('cancel-route-received', (
            json_payload,
            newpacket,
            info,
        ))
        return True

    def _do_connect_proxy_routers_dht_layer(self):
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.dht import known_nodes
        known_seeds = known_nodes.nodes()
        d = dht_service.connect(
            seed_nodes=known_seeds,
            layer_id=dht_records.LAYER_PROXY_ROUTERS,
            attach=True,
        )
        d.addCallback(self._on_dht_proxy_routers_layer_connected)
        d.addErrback(lambda *args: lg.err(str(args)))

    def _on_dht_proxy_routers_layer_connected(self, ok):
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.userid import my_id
        lg.info('connected to DHT layer for proxy routers: %r' % ok)
        if my_id.getIDURL():
            dht_service.set_node_data('idurl', my_id.getIDURL().to_text(), layer_id=dht_records.LAYER_PROXY_ROUTERS)
        return

    def _on_dht_layer_connected(self, evt):
        from bitdust.dht import dht_records
        if evt.data['layer_id'] == 0:
            self._do_connect_proxy_routers_dht_layer()
        elif evt.data['layer_id'] == dht_records.LAYER_PROXY_ROUTERS:
            self._on_dht_proxy_routers_layer_connected(True)



try:
    import Cookie
except:
    import http.cookies as Cookie


class SimpleCookieJar(object):
    def __init__(self):
        self.jar = dict()

    def add(self, set_cookie):
        if set_cookie:
            try:
                simpleCookie = Cookie.SimpleCookie(set_cookie)
            except:
                simpleCookie = Cookie.SimpleCookie(set_cookie.encode('ascii', 'ignore'))

            for k, v in simpleCookie.items():
                domain = v.get('domain')
                if domain:
                    if not domain.startswith('.'):
                        domain = '.' + domain
                    cookie = self.jar.get(domain) if self.jar.get(domain) else Cookie.SimpleCookie()
                    cookie.update(simpleCookie)
                    self.jar[domain.lower()] = cookie

    def set(self, set_cookie):
        if set_cookie:
            try:
                simpleCookie = Cookie.SimpleCookie(set_cookie)
            except:
                simpleCookie = Cookie.SimpleCookie(set_cookie.encode('ascii', 'ignore'))

            for k, v in simpleCookie.items():
                domain = v.get('domain')
                if domain:
                    if not domain.startswith('.'):
                        domain = '.' + domain
                    self.jar[domain.lower()] = simpleCookie

    def get(self, host):
        if not host:
            return ''

        cookies = []
        for domain, simpleCookie in self.jar.items():
            host = host.lower()
            if host.endswith(domain) or host == domain[1:]:
                cookies.append(self.jar.get(domain))

        return '; '.join(filter(None, ['%s=%s' % (k, v.value) for cookie in filter(None, sorted(cookies)) for k, v in sorted(cookie.items())]))
